# encoding: UTF-8
#
# Copyright 2012-2021 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here
from __future__ import unicode_literals
import platform
import logging

try:
    import tkinter as tk
except:
    import Tkinter as tk

from pygubu.builder import builderobject
from .propertieshelp import help_for, _

logger = logging.getLogger(__name__)


TK_BITMAPS = (
    'error', 'gray75', 'gray50', 'gray25', 'gray12',
    'hourglass', 'info', 'questhead', 'question', 'warning',
    )

TK_BITMAPS_MAC = (
    'document', 'stationery', 'edition', 'application', 'accessory',
    'forder', 'pfolder', 'trash', 'floppy', 'ramdisk', 'cdrom',
    'preferences', 'querydoc', 'stop', 'note', 'caution'
    )

TK_CURSORS = (
    'arrow', 'based_arrow_down', 'based_arrow_up', 'boat',
    'bogosity', 'bottom_left_corner', 'bottom_right_corner',
    'bottom_side', 'bottom_tee', 'box_spiral', 'center_ptr',
    'circle', 'clock', 'coffee_mug', 'cross', 'cross_reverse',
    'crosshair', 'diamond_cross', 'dot', 'dotbox', 'double_arrow',
    'draft_large', 'draft_small', 'draped_box', 'exchange', 'fleur',
    'gobbler', 'gumby', 'hand1', 'hand2', 'heart', 'icon',
    'iron_cross', 'left_ptr', 'left_side', 'left_tee', 'leftbutton',
    'll_angle', 'lr_angle', 'man', 'middlebutton', 'mouse', 'none',
    'pencil', 'pirate', 'plus', 'question_arrow', 'right_ptr',
    'right_side', 'right_tee', 'rightbutton', 'rtl_logo',
    'sailboat', 'sb_down_arrow', 'sb_h_double_arrow',
    'sb_left_arrow', 'sb_right_arrow', 'sb_up_arrow',
    'sb_v_double_arrow', 'shuttle', 'sizing', 'spider', 'spraycan',
    'star', 'target', 'tcross', 'top_left_arrow', 'top_left_corner',
    'top_right_corner', 'top_side', 'top_tee', 'trek', 'ul_angle',
    'umbrella', 'ur_angle', 'watch', 'xterm', 'X_cursor')

TK_CURSORS_WINDOWS = (
    'no', 'starting', 'size', 'size_ne_sw'
    'size_ns', 'size_nw_se', 'size_we','uparrow', 'wait'
    )

TK_CURSORS_MAC = (
    'copyarrow', 'aliasarrow', 'contextualmenuarrow', 'text',
    'cross-hair', 'closedhand', 'openhand', 'pointinghand',
    'resizeleft', 'resizeright', 'resizeleftright', 'resizeup',
    'resizedown', 'resizeupdown', 'notallowed', 'poof',
    'countinguphand', 'countingdownhand', 'countingupanddownhand', 'spinning'
    )

if platform.system() == 'Darwin':
    TK_BITMAPS = TK_BITMAPS + TK_BITMAPS_MAC
    TK_CURSORS = TK_CURSORS + TK_CURSORS_MAC
elif platform.system() == 'Windows':
    TK_CURSORS = TK_CURSORS + TK_CURSORS_WINDOWS

TK_RELIEFS = (tk.FLAT, tk.RAISED, tk.SUNKEN, tk.GROOVE, tk.RIDGE)

TK_WIDGET_OPTIONS = {
    'accelerator': {
        'editor': 'entry',
        'help': help_for('accelerator')
        },
    'activerelief': {
        'editor': 'choice',
        'params': {
            'values': ('', tk.FLAT, tk.RAISED, tk.SUNKEN,
                       tk.GROOVE, tk.RIDGE),
            'state': 'readonly'},
        'help': help_for('activerelief')
        },
    'activestyle': {
        'editor': 'choice',
        'params': {
            'values': ('', 'underline', 'dotbox', 'none'),
            'state': 'readonly'},
        'help': help_for('activestyle')
        },
    'activebackground': {
        'editor': 'colorentry',
        'help': help_for('activebackground')
        },
    'activeborderwidth': {
        'editor': 'dimensionentry',
        'help': help_for('activeborderwidth')
        },
    'activeforeground': {
        'editor': 'colorentry',
        'help': help_for('activeforeground')
        },
    'after': {
        'editor': 'entry',
        'help': help_for('after')
        },
    # ttk.Label
    'anchor': {
        'editor': 'choice',
        'params': {
            'values': ('', 'n', 'ne', 'nw', 'e', 'w', 's', 'se', 'sw', 'center'),
            'state': 'readonly'},
        'help': help_for('anchor'),
        'ttk.xxxx?': {
            'params': {
                'values': ('', tk.W, tk.CENTER, tk.E),
                'state': 'readonly'}
            },
        },
    'aspect': {
        'editor': 'naturalnumber',
        'help': help_for('aspect')
        },
    'autoseparators': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('autoseparators')
        },
    # ttk.Label
    'background': {
        'editor': 'colorentry',
        'help': {
            'tk': help_for('background-tk'),
            'ttk': help_for('background-ttk'),
            }
        },
    # ttk.Frame, ttk.Label
    'borderwidth': {
        'editor': 'dimensionentry',
        'help': help_for('borderwidth')
        },
    'bigincrement': {
        'editor': 'naturalnumber',
        'help': help_for('bigincrement')
        },
    'bitmap': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_BITMAPS, 'state': 'readonly'},
        'help': help_for('bitmap')
        },
    'blockcursor': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('blockcursor')
        },
    'buttonbackground': {
        'editor': 'colorentry',
        'help': help_for('buttonbackground'),
        'tk.Spinbox': {
            'help': help_for('buttonbackground-tk.Spinbox')
            }
        },
    'buttoncursor': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_CURSORS, 'state': 'readonly'},
        'help': help_for('buttoncursor')
        },
    'buttondownrelief': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_RELIEFS, 'state': 'readonly'},
        'help': help_for('buttondownrelief')
        },
    'buttonuprelief': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_RELIEFS, 'state': 'readonly'},
        'help': help_for('buttonuprelief')
        },
    'class_': {
        'editor': 'alphanumentry',
        'help': help_for('class_')
        },
    'closeenough': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 999},
        'help': help_for('closeenough')
        },
    'column_anchor': { # ttk.Treeview.Column
        'editor': 'choice',
        'params': {'values': ('', tk.W, tk.CENTER, tk.E), 'state': 'readonly'},
        'default': tk.W,
        'help': help_for('column_anchor')
        },
    'columnbreak': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('columnbreak')
        },
    'command': {
        'editor': 'dynamic',
        'params': {'mode': 'commandentry'},
        'tk.Scrollbar': {
             'params': {'mode': 'scrollcommandentry'}
             },        
        'ttk.Scrollbar': {
             'params': {'mode': 'scrollcommandentry'}
             },
        'ttk.Scale': {
             'params': {'mode': 'scalecommandentry'}
             },
        'tk.Scale': {
             'params': {'mode': 'scalecommandentry'}
             },
        'tk.OptionMenu': {
             'params': {'mode': 'simplecommandentry'}
             },
        'help': help_for('command-pygubu')
        },
    # ttk.Label
    'compound': {
        'editor': 'choice',
        'params': {
            'values': ('', tk.TOP, tk.BOTTOM, tk.LEFT, tk.RIGHT, tk.CENTER, tk.NONE),
            'state': 'readonly'},
        'help': {
            'tk': help_for('compound-tk'),
            'ttk': help_for('compound-ttk')}
        },
    # ttk.Button
    'confine': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('confine')
        },
    'container': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('container')
        },
    'cursor': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_CURSORS, 'state': 'readonly'},
        'help': help_for('cursor')
        },
    # ttk.Button
    'default': {
        'editor': 'choice',
        'params': {
            'values': ('', 'normal', 'active', 'disabled'),
            'state': 'readonly'},
        'help': help_for('default')
        },
    'digits': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 999},
        'help': help_for('digits')
        },
    'direction': {
        'editor': 'choice',
        'help': help_for('direction'),
        'tk.Menubutton': {
            'params': {'values': ('', tk.LEFT, tk.RIGHT, 'above'),
                       'state': 'readonly'}},
        'ttk.Menubutton': {
            'params': {
                'values': ('', 'above', 'below', 'flush',
                           tk.LEFT, tk.RIGHT),
                'state': 'readonly'}},
        },
    'disabledbackground': {
        'editor': 'colorentry',
        'help': help_for('disabledbackground')
        },
    'disabledforeground': {
        'editor': 'colorentry',
        'help': help_for('disabledforeground')
        },
    'elementborderwidth': {
        'editor': 'dimensionentry',
        'help': help_for('elementborderwidth')
        },
    'endline': {
        'editor': 'naturalnumber',
        'help': help_for('endline')
        },
    # ttk.Checkbutton, ttk.Entry
    'exportselection': {
        'editor': 'choice',
        'params': {'values': ('', 'true', 'false'), 'state': 'readonly'},
        'help': help_for('exportselection')
        },
    # ttk.Label
    'font': {
        'editor': 'fontentry',
        'help': help_for('font')
        },
    # ttk.Label
    'foreground': {
        'editor': 'colorentry',
        'help': {
            'tk': help_for('foreground-tk'),
            'ttk': help_for('foreground-ttk')
            }
        },
    # ttk.Spinbox
    'format': {
        'editor': 'entry'
        },
    # ttk.Scale, ttk.Spinbox
    'from_': {
        'editor': 'realnumber',
        'help': help_for('from_')
        },
    'handlepad': {
        'editor': 'dimensionentry',
        'help': help_for('handlepad')
        },
    'handlesize': {
        'editor': 'dimensionentry',
        'help': help_for('handlesize')
        },
    # ttk.Treeview.Column
    'heading_anchor': {
        'editor': 'choice',
        'params': {
            'values': ('', tk.W, tk.CENTER, tk.E), 'state': 'readonly'},
        'default': tk.W,
        'help': help_for('heading_anchor')
        },
    # ttk.Frame,
    'height': {
        'editor': 'dynamic',
        'help': help_for('height'),
        'params': {'mode': 'dimensionentry'},
        'tk.Button': {
            'help': help_for('height-tk.Button')
            },
        'ttk.Combobox': {
            'params': {'mode': 'naturalnumber'},
            'help': help_for('height-ttk.Combobox')
            },
        'tk.Toplevel': {'default': 200},
        'tk.Frame': {'default': 200},
        'ttk.Frame': {'default': 200},
        'tk.LabelFrame': {'default': 200},
        'ttk.Labelframe': {'default': 200},
        'tk.Listbox': {
            'params': {'mode': 'naturalnumber'},
            },
        'tk.Menubutton': {
            'params': {'mode': 'naturalnumber'},
            'help': help_for('height-tk.Menubutton')
            },        
        'tk.PanedWindow': {'default': 200},
        'ttk.Panedwindow': {'default': 200},
        'ttk.Notebook': {'default': 200},
        'tk.Text': {
            'params': {'mode': 'naturalnumber'},
            'default': 10,
            'help': help_for('height-tk.Text')
            },
        'tk.Radiobutton':{
            'params': {'mode': 'naturalnumber'},
            },
        'ttk.Treeview': {
            'params': {'mode': 'naturalnumber'},
            'help': help_for('height-ttk.Treeview')
            },
        'pygubu.builder.widgets.editabletreeview': {
            'params': {'mode': 'naturalnumber'},
            'help': help_for('height-ttk.Treeview')
            },
        'pygubu.builder.widgets.dialog': {'default': 100}
        },
    'hidemargin': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('hidemargin')
        },
    'highlightbackground': {
        'editor': 'colorentry',
        'help': help_for('highlightbackground')
        },
    'highlightcolor': {
        'editor': 'colorentry',
        'help': help_for('highlightcolor')
        },
    'highlightthickness': {
        'editor': 'dimensionentry',
        'help': help_for('highlightthickness')
        },
    # ttk.Label
    'image': {
        'editor': 'imageentry',
        'help': {
            'tk': help_for('image-tk'),
            'ttk': help_for('image-ttk')
            },
        'ttk.Treeview.Column': {
            'help': help_for('image-ttk.Treeview.Column'),
            }
        },
    'inactiveselectbackground': {
        'editor': 'colorentry',
        'help': help_for('inactiveselectbackground')
        },
    # ttk.Spinbox
    'increment': {
        'editor': 'realnumber',
        'help': help_for('increment')
        },
    'indicatoron': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('indicatoron')
        },
    'insertbackground': {
        'editor': 'colorentry',
        'help': help_for('insertbackground')
        },
    'insertborderwidth': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 999},
        'help': help_for('insertborderwidth')
        },
    'insertofftime': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 9999, 'increment': 100},
        'help': help_for('insertofftime')
        },
    'insertontime': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 9999, 'increment': 100},
        'help': help_for('insertontime')
        },
    'insertunfocussed': {
        'editor': 'choice',
        'params': {
            'values': ('', 'none', 'hollow', 'solid'),
            'state': 'readonly'},
        'help': help_for('insertunfocussed')
        },
    'insertwidth': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 999},
        'help': help_for('insertwidth')
        },
    # ttk.Entry
    'invalidcommand': {
        'editor': 'entryvalidatecommandentry',
        'help': help_for('command-pygubu')
        },
    'jump': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('jump')
        },
    # ttk.Label
    'justify': {
        'editor': 'choice',
        'params': {
            'values': ('', 'left', 'center', 'right'), 'state': 'readonly'},
        'help': help_for('justify'),
        'ttk.Entry': {
            'help': help_for('justify-ttk.Entry')}
        },
    'label': {
        'editor': 'entry',
        'help': help_for('label')
        },
    # ttk.Labelframe
    'labelanchor': {
        'editor': 'choice',
        'params': {
            'values': ('', 'nw', 'n', 'ne', 'en', 'e', 'es',
                       'se', 's', 'sw', 'ws', 'w'),
            'state': 'readonly'},
        'help': help_for('labelanchor')
        },
    # ttk.Progressbar
    'length': {
        'editor': 'dimensionentry',
        'help': help_for('length')
        },
    'listvariable': {
        'editor': 'tkvarentry',
        'help': help_for('listvariable')
        },
    # ttk.Progressbar
    'maximum': {
        'editor': 'realnumber',
        'help': help_for('maximum')
        },
    'maxundo': {
        'editor': 'integernumber',
        'help': help_for('maxundo')
        },
    'minsize': {
        'editor': 'dimensionentry',
        'help': help_for('minsize')
        },
    # ttk.Treeview.Column
    'minwidth': {
        'editor': 'naturalnumber',
        'default': '20',
        'help': help_for('minwidth')
        },
    # ttk.Progressbar
    'mode': {
        'editor': 'choice',
        'params': {
            'values': ('', 'determinate', 'indeterminate'),
            'state': 'readonly'}},
    'offrelief': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_RELIEFS, 'state': 'readonly'},
        'help': help_for('offrelief')
        },
    # ttk.Checkbutton
    'offvalue': {
        'editor': 'entry',
        'help': help_for('offvalue')
        },
    # ttk.Checkbutton
    'onvalue': {
        'editor': 'entry',
        'help': help_for('onvalue')
        },
    'opaqueresize': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('opaqueresize'),
        },
    # ttk.Panedwindow
    'orient': {
        'editor': 'choice',
        'params': {'values': (tk.VERTICAL, tk.HORIZONTAL),
                   'state': 'readonly'},
        'default': tk.HORIZONTAL,
        'help': help_for('orient')
        },
    'overrelief': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_RELIEFS, 'state': 'readonly'},
        'help': help_for('overrelief')
        },
    # ttk.Frame, ttk.Label
    'padding': {
        'editor': 'fourdimensionentry',
        'help': help_for('padding')
        },
    'padx': {
        'editor': 'dimensionentry',
        'help': help_for('padx')
        },
    'pady': {
        'editor': 'dimensionentry',
        'help': help_for('pady')
        },
    # ttk.Checkbutton
    'postcommand': {
        'editor': 'simplecommandentry',
        'help': help_for('postcommand')
        },
    'proxybackground': {
        'editor': 'colorentry',
        'help': help_for('proxybackground')
        },
    'proxyborderwidth': {
        'editor': 'dimensionentry',
        'help': help_for('proxyborderwidth')
        },
    'proxyrelief': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_RELIEFS, 'state': 'readonly'},
        'help': help_for('proxyrelief')
        },
    'readonlybackground': {
        'editor': 'colorentry',
        'help': help_for('readonlybackground')
        },
    # ttk.Frame,
    'relief': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_RELIEFS, 'state': 'readonly'},
        'help': help_for('relief')
        },
    'repeatdelay': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 9999, 'increment': 100},
        'help': help_for('repeatdelay')
        },
    'repeatinterval': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 9999, 'increment': 100},
        'help': help_for('repeatinterval')
        },
    'resolution': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 999, 'increment': 0.5},
        'help': help_for('resolution')
        },
    'sliderlength': {
        'editor': 'dimensionentry',
        'help': help_for('sliderlength')
        },
    'sliderrelief': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_RELIEFS, 'state': 'readonly'},
        'help': help_for('sliderrelief')
        },
    'sashcursor': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_CURSORS, 'state': 'readonly'},
        'help': help_for('sashcursor')
        },
    'sashpad': {
        'editor': 'dimensionentry',
        'help': help_for('sashpad')
        },
    'sashrelief': {
        'editor': 'choice',
        'params': {'values': ('',) + TK_RELIEFS, 'state': 'readonly'},
        'help': help_for('sashrelief')
        },
    'sashwidth': {
        'editor': 'dimensionentry',
        'help': help_for('sashwidth')
        },
    'selectbackground': {
        'editor': 'colorentry',
        'help': help_for('selectbackground')
        },
    'selectborderwidth': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 999},
        'help': help_for('selectborderwidth')
        },
    'selectforeground': {
        'editor': 'colorentry',
        'help': help_for('selectforeground')
        },
    'scrollregion': {
        'editor': 'fourdimensionentry',
        'help': help_for('scrollregion')
        },
    'selectcolor': {
        'editor': 'colorentry',
        'help': help_for('selectcolor')
        },
    'selectimage': {
        'editor': 'imageentry',
        'help': help_for('selectimage')
        },
    # ttk.Treeview
    'selectmode': {
        'editor': 'choice',
        'help': help_for('selectmode'),
        'params': {
            'values': ('', tk.BROWSE, tk.SINGLE, tk.MULTIPLE, tk.EXTENDED),
            'state': 'readonly'},
        'ttk.Treeview': {
            'params': {
                'values': (tk.EXTENDED, tk.BROWSE, tk.NONE),
                'state': 'readonly'},
            'default': tk.EXTENDED,
            }
        },
    'setgrid': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('setgrid')
        },
    # ttk.Entry
    'show': {
        'editor': 'choice',
        'params': {'values': ('', '•'), 'state': 'normal'},
        'help': help_for('show'),
        'ttk.Treeview': {
            'params': {
                'values': ('', 'tree', 'headings'), 'state': 'readonly'},
            'help': help_for('show-ttk.Treeview')
            },
        'pygubu.builder.widgets.editabletreeview': {
            'params': {
                'values': ('', 'tree', 'headings'), 'state': 'readonly'},
            'help': help_for('show-ttk.Treeview')
            },
        },
    'showhandle': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('showhandle')
        },
    'showvalue': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('showvalue')
        },
    'spacing1': {
        'editor': 'dimensionentry',
        'help': help_for('spacing1')
        },
    'spacing2': {
        'editor': 'dimensionentry',
        'help': help_for('spacing2')
        },
    'spacing3': {
        'editor': 'dimensionentry',
        'help': help_for('spacing3')
        },
    'startline': {
        'editor': 'naturalnumber',
        'help': help_for('startline')
        },
    'state': {
        'editor': 'choice',
        'help': help_for('state'),
        'params': {'values': ('', tk.NORMAL, tk.DISABLED),
                   'state': 'readonly'},
        'tk.Button': {
            'params': {
                'values': ('', tk.NORMAL, tk.ACTIVE, tk.DISABLED),
                'state': 'readonly'}},
        'tk.Checkbutton': {
            'params': {
                'values': ('', tk.NORMAL, tk.ACTIVE, tk.DISABLED),
                'state': 'readonly'},
                'help': help_for('state-tk.Checkbutton')
            },
        'tk.Canvas': {
            'params': {
                'values': ('', tk.NORMAL, tk.DISABLED),
                'state': 'readonly'},
            'help': help_for('state-tk.Canvas')
            },
        'tk.Entry': {
            'params': {
                'values': ('', tk.NORMAL, tk.DISABLED, 'readonly'),
                'state': 'readonly'}},
        'tk.Combobox': {
            'params': {
                'values': ('', 'readonly'), 'state': 'readonly'}},
        'ttk.Entry': {
            'params': {
                'values': ('', tk.NORMAL, tk.DISABLED, 'readonly'),
                'state': 'readonly'}},
        'ttk.Combobox': {
            'params': {
                'values': ('', 'normal', 'readonly', 'disabled'),
                'state': 'readonly'}},
        'ttk.Button': {
            'params': {
                'values': ('', 'normal', 'disabled'),
                'state': 'readonly'}},
        'ttk.Notebook.Tab': {
            'params': {
                'values': ('', 'normal', 'disabled', 'hidden'),
                'state': 'readonly'
                }},
        'tk.Spinbox': {
            'params': {
                'values': ('', tk.NORMAL, tk.DISABLED, 'readonly'),
                'state': 'readonly'
                }
            },
    },
    # ttk.Notebook.Tab
    'sticky': {
        'editor': 'stickyentry',
        'params': {},
        'help': help_for('sticky')
        },
    # ttk.Treeview.Column
    'stretch': {
        'editor': 'choice',
        'help': help_for('stretch'),
        'ttk.Treeview.Column': {
            'params': {'values': ('true', 'false'), 'state': 'readonly'},
            'default': 'true',
            'help': help_for('stretch-ttk.Treeview')
            },
        'tk.PanedWindow.Pane': {
            'params': {
                'values': ('', 'always', 'first', 'last', 'middle', 'never'),
                'state': 'readonly'}}
        },
    'style': {
        'editor': 'ttkstylechoice',
        'ttk.Button': {
            'params': {'values': ('', 'Toolbutton')}
            },
        'help': help_for('style')
        },
    'tabs': { # FIXME see tk.Text tab property
        'editor': 'entry',
        'help': help_for('tabs')
        },
    'tabstyle': {
        'editor': 'choice',
        'params': {
            'values': ('', 'tabular', 'wordprocessor'),
            'state': 'readonly'},
        'help': help_for('tabstyle')
        },
    'takefocus': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('takefocus')
        },
    'tearoff': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('tearoff')
        },
    'tearoffcommand': {
        'editor': 'simplecommandentry',
        'help': help_for('tearoffcommand')
        },
    # ttk.Label
    'text': {
        'editor': 'text',
        'help': help_for('text')
        },
    # ttk.Label
    'textvariable': {
        'editor': 'tkvarentry',
        'help': help_for('textvariable')
        },
    'tickinterval': {
        'editor': 'spinbox',
        'params': {'from_': 0, 'to': 999, 'increment': 0.5},
        'help': help_for('tickinterval')
        },
    # ttk.Scale, ttk.Spinbox
    'to': {
        'editor': 'realnumber',
        'help': help_for('to')
        },
    'title': {
        'editor': 'entry',
        'help': help_for('title-menu')
        },
    'tristateimage': {
        'editor': 'imageentry',
        'help': help_for('tristateimage')
        },
    'tristatevalue': {
        'editor': 'entry',
        'help': help_for('tristatevalue')
        },
    'troughcolor': {
        'editor': 'colorentry',
        'help': help_for('troughcolor')
        },
    # ttk.Label
    'underline': {
        'editor': 'naturalnumber',
        'help': help_for('underline')
        },
    'undo': {
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('undo')
        },
    'value': {
        'editor': 'dynamic',
        'help': help_for('value'),
        'params': {'mode': 'entry'},
        'ttk.Progressbar':{
            'params':{'mode':'realnumber'}
            },
        'ttk.Scale':{
            'params':{'mode':'realnumber'}
            },
        },
    # ttk.Checkbutton
    'values': {
        'editor': 'entry',
        'help': help_for('values'),
        'tk.Spinbox': {
            'help': help_for('values-tk.Spinbox')
            }
        },
    'validate': {
        'editor': 'choice',
        'params': {
            'values': ('', 'none', 'focus', 'focusin',
                       'focusout', 'key', 'all'),
            'state': 'readonly'},
        'help': help_for('validate')
        },
    'validatecommand': {
        'editor': 'entryvalidatecommandentry',
        'help': help_for('command-pygubu')
        },
    # ttk.Checkbutton
    'variable': {
        'editor': 'tkvarentry',
        'help': help_for('variable')
        },
    # ttk.Panedwindow.Pane
    'weight': {
        'editor': 'naturalnumber',
        'default': '1',
        'help': help_for('weight')
        },
    # ttk.Frame, ttk.Label
    'width': {
        'editor': 'dynamic',
        'help': help_for('width'),
        'params': {'mode': 'dimensionentry'},
        'tk.Label': {
            'help': help_for('width-tk.Label')
            },
        'tk.Message': {
            'help': help_for('width-tk.Message')
            },        
        'tk.Button': {
            'params': {'mode': 'integernumber'},
            'help': help_for('width-tk.Button')
            },
        'tk.Scale': {
            'help': help_for('width-scale')
            },
        'ttk.Button': {
            'params': {'mode': 'integernumber'},
            'help': help_for('width-ttk')
            },
        'tk.Canvas': {
            'params': {'mode': 'entry'}
            },
        'tk.Checkbutton': {
            'params': {'mode': 'integernumber'},
            'help': help_for('width-tk')
            },
        'tk.Spinbox': {
            'params': {'mode': 'integernumber'},
            'help': help_for('width-tk.Spinbox')
            },
        'tk.Scrollbar': {
            'params': {'mode': 'dimensionentry'},
            'help': help_for('width-tk.Scrollbar')
            },
        'ttk.Checkbutton': {
            'params': {'mode': 'integernumber'},
            'help': help_for('width-ttk')
            },
        'tk.Entry': {
            'params': {'mode': 'naturalnumber'},
            'help': help_for('width-tk.Entry')
            },
        'ttk.Entry': {
            'params': {'mode': 'naturalnumber'},
            },
        'tk.Menubutton': {
            'params': {'mode': 'naturalnumber'},
            'help': help_for('width-tk.Menubutton')
            },
        'ttk.Label': {
            'params': {'mode': 'integernumber'},
            'help': help_for('width-ttk.Label')
            },
        'tk.Listbox': {
            'params': {'mode': 'naturalnumber'},
            },
        'tk.Frame': {
            'default': 200
            },
        'ttk.Frame': {
            'default': 200},
        'tk.LabelFrame': {
            'default': 200},
        'ttk.Labelframe': {
            'default': 200},
        'tk.PanedWindow': {
            'default': 200},
        'ttk.Panedwindow': {
            'default': 200},
        'ttk.Notebook': {
            'default': 200},
        'tk.Radiobutton':{
            'params': {'mode': 'naturalnumber'},
            },
        'ttk.Radiobutton':{
            'params': {'mode': 'integernumber'},
            'help': help_for('width-ttk')
            },
        'tk.Text': {
            'params': {'mode': 'naturalnumber'},
            'default': 50,
            'help': help_for('width-tk.Text')
            },
        'tk.Toplevel': {
            'default': 200},
        'ttk.Treeview.Column': {
            'params': {'mode': 'naturalnumber'},
            'default': 200,
            'help': help_for('width-ttk.Treeview.Column')
            },
        'pygubu.builder.widgets.dialog': {
            'default': 200}},
    # ttk.Spinbox
    'wrap': {
        'editor': 'choice',
        'params': {
            'values': ('', 'false', 'true'),
            'state': 'readonly'},
        'help': help_for('wrap-ttk.Spinbox'),
        'tk.Text': {
            'params': {
                'values': ('', tk.CHAR, tk.WORD, tk.NONE),
                'state': 'readonly'},
            'help': help_for('wrap-tk.Text')
            }
        },
    # ttk.Label
    'wraplength': {
        'editor': 'dimensionentry',
        'help': help_for('wraplength')
        },
    # ttk.Entry
    'xscrollcommand': {
        'editor': 'scrollsetcommandentry',
        'help': help_for('command-pygubu')
        },
    'xscrollincrement': {
        'editor': 'dimensionentry',
        'help': help_for('xscrollincrement')
        },
    # ttk.Treeview
    'yscrollcommand': {
        'editor': 'scrollsetcommandentry',
        'help': help_for('command-pygubu')
        },
    'yscrollincrement': {
        'editor': 'dimensionentry',
        'help': help_for('yscrollincrement')
        },
    }

REQUIRED_OPTIONS = {
    'class': {
        'editor': 'entry',
        'params': {'state': 'readonly'},
        'help': help_for('class')
    },
    'id': {
        'editor': 'identifierentry',
        'help': help_for('id')
    },
}

CUSTOM_OPTIONS = {
    'geometry': {
        'editor': 'dynamic',
        'help': help_for('geometry-custom'),
        'params': {
            'mode': 'geometryentry',
            'values': ('', '320x200', '320x240', '352x288', '384x288',
                       '480x320', '640x480', '768x576', '800x480', '800x600',
                       '854x480', '1024x576', '1024x600', '1024x768',
                       '1152x768', '1152x864', '1280x720', '1280x768',
                       '1280x800', '1280x854', '1280x960', '1280x1024',
                       '1366x768', '1400x1050', '1440x1080', '1440x900',
                       '1440x960', '1440x1080', '1600x900', '1600x1050',
                       '1600x1200', '1920x1080', '2048x1080'),
            },
        },
    'iconbitmap': {
        'editor': 'dynamic',
        'params': {
            'mode': 'imageentry',},
        'help': help_for('iconbitmap-custom')
        },
    'iconphoto': {
        'editor': 'dynamic',
        'params': {
            'mode': 'imageentry',},
        'help': help_for('iconphoto-custom')
        },
    'maxsize': {
        'editor': 'whentry',
        'help': help_for('maxsize-custom')
        },
    'minsize': {
        'editor': 'whentry',
        'help': help_for('minsize-custom')
        },
    'overrideredirect': {
        'editor': 'dynamic',
        'params': {
            'mode': 'choice',
            'values': ('', 'True', 'False'),
            'state': 'readonly'
            },
        'help': help_for('overrideredirect-custom')
        },
    'resizable': {
        'editor': 'dynamic',
        'params': {
            'mode': 'choice',
            'values': ('', 'both', 'horizontally', 'vertically', 'none'),
            'state': 'readonly'},
        'help': help_for('resizable-custom')
        },
    'scrolltype': {
        'editor': 'dynamic',
        'params': {
            'mode': 'choice',
            'values': ('both', 'vertical', 'horizontal'),
            'state': 'readonly'},
        'default': 'both',
        'help': help_for('scrolltype-custom')
        },
    'text': {
        'editor': 'dynamic',
        'params': {
            'mode': 'text',},
        'help': help_for('text-custom')
        },
    'title': {
        'editor': 'dynamic',
        'params': {
            'mode': 'entry',},
        'help': help_for('title-custom')
        },
    'tree_column': {
        'editor': 'dynamic',
        'params': {
            'mode': 'choice',
            'values': ('true', 'false'),
            'state': 'readonly'},
        'default': 'false',
        'help': help_for('tree_column-custom')
        },
    'usemousewheel': {
        'editor': 'dynamic',
        'params': {
            'mode': 'choice',
            'values': ('true', 'false'),
            'state': 'readonly'},
        'default': 'false'},
    'visible': {
        'editor': 'dynamic',
        'params': {
            'mode': 'choice',
            'values': ('true', 'false'),
            'state': 'readonly'},
        'default': 'true',
        'help': help_for('visible-custom')
        },
    'specialmenu': {
        'editor': 'dynamic',
        'params': {
            'mode': 'choice',
            'values': ('apple', 'help', 'window', 'system'),
            'state': 'readonly'}
        },
    }

WIDGET_REQUIRED_OPTIONS = ('class', 'id')
WIDGET_STANDARD_OPTIONS = (
    'activerelief', 'activestyle', 'activebackground',
    'activeborderwidth', 'activeforeground', 'after',
    'anchor', 'background', 'bitmap', 'borderwidth',
    'class_', 'compound', 'cursor', 'disabledforeground',
    'exportselection',
    'font', 'foreground', 'jump', 'highlightbackground',
    'highlightcolor', 'highlightthickness', 'image',
    'indicatoron', 'insertbackground',
    'insertborderwidth', 'insertofftime', 'insertontime', 'insertwidth',
    'justify', 'orient', 'padx', 'pady', 'relief',
    'repeatdelay', 'repeatinterval', 'selectbackground', 'selectborderwidth',
    'selectforeground', 'setgrid', 'state', 'style', 'takefocus', 'text',
    'textvariable', 'troughcolor', 'underline', 'width', 'wraplength',
    'xscrollcommand', 'yscrollcommand')

WIDGET_SPECIFIC_OPTIONS = (
    'accelerator', 'activestyle', 'activerelief', 'anchor', 'aspect',
    'autoseparators', 'background', 'bigincrement',
    'blockcursor', 'borderwidth', 'buttonbackground', 'buttoncursor',
    'buttondownrelief', 'buttonuprelief',
    'class_', 'column_anchor', 'columnbreak', 'command', 'compound', 'container',
    'closeenough', 'confine', 'default', 'digits', 'direction',
    'disabledbackground', 'disabledforeground', 'elementborderwidth',
    'endline', 'exportselection', 'font',
    'foreground', 'format', 'from_', 'to',
    'handlepad', 'handlesize', 'hidemargin', 'heading_anchor', 'height', 'image',
    'inactiveselectbackground', 'increment', 'indicatoron', 'insertunfocussed',
    'invalidcommand', 'justify',
    'label', 'labelanchor', 'listvariable', 'length',
    'maximum', 'maxundo',
    'minsize', 'minwidth', 'mode', 'offrelief', 'offvalue',
    'onvalue', 'opaqueresize', 'orient', 'overrelief',
    'padding', 'padx', 'pady',
    'postcommand', 'proxybackground', 'proxyborderwidth', 'proxyrelief',
    'readonlybackground', 'relief', 'resolution',
    'scrollregion', 'sashcursor', 'sashpad', 'sashrelief', 'sashwidth',
    'selectcolor', 'selectimage', 'selectmode', 'show',
    'showhandle', 'showvalue', 'sliderlength', 'sliderrelief',
    'spacing1', 'spacing2', 'spacing3', 'startline',
    'state', 'sticky', 'stretch', 'tabs', 'tabstyle',
    'text', 'textvariable', 'title', 'tickinterval', 'tristateimage',
    'tristatevalue', 'underline', 'validate', 'undo', 'validatecommand',
    'value', 'values', 'variable', 'weight', 'width', 'wrap',
    'wraplength', 'xscrollcommand', 'xscrollincrement', 'yscrollincrement',
    'tearoff', 'tearoffcommand'
    )

WIDGET_CUSTOM_OPTIONS = [
    'tree_column', 'visible', 'scrolltype', 'text', 'title',
    'geometry', 'overrideredirect', 'resizable', 'minsize',
    'maxsize', 'usemousewheel', 'iconbitmap', 'iconphoto',
    'specialmenu', 
    ]

WIDGET_PROPERTIES = wp = dict(TK_WIDGET_OPTIONS)
wp.update(REQUIRED_OPTIONS)
wp.update(CUSTOM_OPTIONS)

LAYOUT_OPTIONS = {
    # pack/grid/place common properties
    'padx': {
        'editor': 'twodimensionentry',
        'help': help_for('padx-layout')
        },
    'pady': {
        'editor': 'twodimensionentry',
        'help': help_for('pady-layout')
        },
    'ipadx': {
        'editor': 'twodimensionentry',
        'help': help_for('ipadx-layout')
        },
    'ipady': {
        'editor': 'twodimensionentry',
        'help': help_for('ipady-layout')
        },
    'propagate': {
        'editor': 'choice',
        'params': {'values': ('True', 'False'), 'state': 'readonly'},
        'default': 'True',
        'help': help_for('propagate-layout')
        },
    'anchor': {
        'editor': 'choice',
        'params': {
            'values': ('', 'n', 'ne', 'nw', 'e', 'w', 's', 'se', 'sw', 'center'),
            'state': 'readonly'
            },
        'help': help_for('anchor-layout'),
        'place': {
            'default': 'nw',
            'help': help_for('anchor-layout-place')
            }
        },
    
    # pack properties
    'expand':{
        'editor': 'choice',
        'params': {'values': ('', 'false', 'true'), 'state': 'readonly'},
        'help': help_for('expand-pack')
        },
    'fill':{
        'editor': 'choice',
        'params': {'values': ('', 'x', 'y', 'both'), 'state': 'readonly'},
        'help': help_for('fill-pack')
        },
    'side':{
        'editor': 'choice',
        'params': {
            'values': ('top', 'bottom', 'left', 'right'),
            'state': 'readonly'
            },
        'default': 'top',
        'help': help_for('side-pack')
        },
    'bordermode':{
        'editor': 'choice',
        'params':{
            'values': ('', 'inside', 'outside', 'ignore'),
            'state': 'readonly'},
        'help': help_for('bordermode-pack')
        },
    'height':{
        'editor': 'pixelcoordinateentry',
        'help': help_for('height-pack')
        },
    'relheight':{
        'editor': 'relativeentry',
        'help': help_for('relheight-pack')
        },
    'relwidth':{
        'editor': 'relativeentry',
        'help': help_for('relwidth-pack')
        },
    'relx':{
        'editor': 'relativeentry',
        'help': help_for('relx-pack')
        },
    'rely':{
        'editor': 'relativeentry',
        'help': help_for('rely-pack')
        },
    'width':{
        'editor': 'pixelcoordinateentry',
        'help': help_for('width-pack')
        },
    'x':{
        'editor': 'pixelcoordinateentry',
        'default': '0',
        'help': help_for('x-pack')
        },
    'y':{
        'editor': 'pixelcoordinateentry',
        'default': '0',
        'help': help_for('y-pack')
        },
    #
    # grid packing properties
    #
    'row': {
        'editor': 'naturalnumber',
        'default': '0',
        'help': help_for('row-grid')
        },
    'column': {
        'editor': 'naturalnumber',
        'default': '0',
        'help': help_for('column-grid')
        },
    'sticky': {
        'editor': 'stickyentry',
        'params': {},
        'help': help_for('sticky-grid')
        },
    'rowspan': {
        'editor': 'naturalnumber',
        'help': help_for('rowspan-grid')
        },
    'columnspan': {
        'editor': 'naturalnumber',
        'help': help_for('columnspan-grid')
        },
    #
    # grid row and column properties (can be applied to each row or column)
    'minsize': {
        'editor': 'dimensionentry',
        'params': {'width': 4, 'empty_data': 0},
        'help': help_for('minsize-grid')
        },
    'pad': {
        'editor': 'dimensionentry',
        'params': {'width': 4, 'empty_data': 0},
        'help': help_for('pad-grid')
        },
    'weight': {
        'editor': 'naturalnumber',
        'help': help_for('weight-grid')
    },
    'uniform': {
        'editor': 'alphanumentry',
        'help': help_for('uniform-grid')
    }
}

# List properties in display order
MANAGER_PROPERTIES = (
    'anchor', 
    'relx', 'rely', 'relwidth', 'relheight',
    'x', 'y', 'width', 'height', 'bordermode',
    'side', 'expand', 'fill',
    'row', 'column', 'sticky', 'rowspan', 'columnspan', 'padx', 'pady',
    'ipadx', 'ipady', 'propagate'
)

GRID_PROPERTIES = (
    'row', 'column', 'sticky', 'rowspan', 'columnspan', 'padx', 'pady',
    'ipadx', 'ipady', 'propagate')

PACK_PROPERTIES = (
    'anchor', 'side', 'expand', 'fill', 'padx', 'pady',
    'ipadx', 'ipady', 'propagate')

PLACE_PROPERTIES = (
    'anchor', 'relx', 'rely', 'relwidth', 'relheight',
    'x', 'y', 'width', 'height', 'bordermode', 
)

GRID_RC_PROPERTIES = ('minsize', 'pad', 'weight', 'uniform')

TRANSLATABLE_PROPERTIES = (
    'label', 'text', 'title',
)


def _register_custom(name, descr):
    if name in CUSTOM_OPTIONS:
        CUSTOM_OPTIONS[name].update(descr)
        logger.debug('Updated property: %s', name)
    else:
        CUSTOM_OPTIONS[name] = descr
        WIDGET_CUSTOM_OPTIONS.append(name)
        WIDGET_CUSTOM_OPTIONS.sort()
        logger.debug('Registered property: %s', name)
    WIDGET_PROPERTIES.update(CUSTOM_OPTIONS)

def register_property(name, descr):
    _register_custom(name, descr)
    builderobject._old_register_property(name, descr)

if not hasattr(builderobject, '_register_fixed_'):
    for name, descr in builderobject.CUSTOM_PROPERTIES.items():
        _register_custom(name, descr)
    builderobject._register_fixed_ = True
    builderobject._old_register_property = builderobject.register_property
    builderobject.register_property = register_property
    logger.debug('Installed custom register_property function')



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import os
import xml.etree.ElementTree as ET
from collections import Counter
import logging
try:
    import tkinter as tk
    from tkinter import messagebox
except:
    import Tkinter as tk
    import tkMessageBox as messagebox

from pygubu.builder import CLASS_MAP
from pygubu.builder.uidefinition import UIDefinition
from pygubu.stockimage import StockImage, StockImageException
import pygubudesigner
from .widgetdescr import WidgetMeta
from .i18n import translator as _
from .util import trlog
from .propertieseditor import PropertiesEditor
from .bindingseditor import BindingsEditor
from .layouteditor import LayoutEditor
from .i18n import translator
from .actions import *
from pygubudesigner import preferences as pref


logger = logging.getLogger('pygubu.designer')

#translator function
_ = translator


class WidgetsTreeEditor(object):
    GRID_UP = 0
    GRID_DOWN = 1
    GRID_LEFT = 2
    GRID_RIGHT = 3

    def __init__(self, app):
        self.app = app
        self.treeview = app.treeview
        self.previewer = app.previewer
        self.treedata = {}
        self.counter = Counter()
        self.default_layout_manager = 'pack' # TODO: set from configuration
        # Filter vars
        self.filter_on = False
        self.filtervar = app.builder.get_variable('filtervar')
        self.filter_btn = app.builder.get_object('filterclear_btn')
        self.filter_prev_value = ''
        self.filter_prev_sitem = None
        self._detached = []
        self._listen_object_updates = True

        self.config_treeview()
        self.config_filter()
        
        # current item being edited
        self.current_edit = None

        # Widget Editor
        pframe = app.builder.get_object('propertiesframe')
        lframe = app.builder.get_object('layoutframe')
        bframe = app.builder.get_object('bindingsframe')
        bindingstree = app.builder.get_object('bindingstree')
        
        self.properties_editor = PropertiesEditor(
            pframe, id_validator=self.is_id_unique)
        self.layout_editor = LayoutEditor(lframe)
        self.bindings_editor = BindingsEditor(bindingstree, bframe)
        self.treeview.bind_all('<<PreviewItemSelected>>', self._on_preview_item_clicked)
        # Listen to Grid RC changes from layout
        lframe.bind_all('<<LayoutEditorGridRCChanged>>', self._on_gridrc_changed)
        f = lambda e, manager='grid': self.change_container_manager(manager)
        lframe.bind_all('<<LayoutEditorContainerManagerToGrid>>', f)
        f = lambda e, manager='pack': self.change_container_manager(manager)
        lframe.bind_all('<<LayoutEditorContainerManagerToPack>>', f)
        
        # Tree Editing
        tree = self.treeview
        tree.bind_all(TREE_ITEM_COPY, lambda e: self.copy_to_clipboard())
        tree.bind_all(TREE_ITEM_PASTE, lambda e: self.paste_from_clipboard())
        tree.bind_all(TREE_ITEM_CUT, lambda e: self.cut_to_clipboard())
        tree.bind_all(TREE_ITEM_DELETE, self.on_tree_item_delete)
        tree.bind_all(TREE_ITEM_GRID_DOWN,
                  lambda e: self.on_item_grid_move(self.GRID_DOWN))
        tree.bind_all(TREE_ITEM_GRID_LEFT,
                  lambda e: self.on_item_grid_move(self.GRID_LEFT))
        tree.bind_all(TREE_ITEM_GRID_RIGHT,
                  lambda e: self.on_item_grid_move(self.GRID_RIGHT))
        tree.bind_all(TREE_ITEM_GRID_UP,
                  lambda e: self.on_item_grid_move(self.GRID_UP))
        tree.bind_all(TREE_ITEM_MOVE_UP, self.on_item_move_up)
        tree.bind_all(TREE_ITEM_MOVE_DOWN, self.on_item_move_down)
        tree.bind_all(TREE_NAV_UP, self.on_item_nav_up)
        tree.bind_all(TREE_NAV_DOWN, self.on_item_nav_down)
        tree.bind_all(TREE_ITEM_PREVIEW_TOPLEVEL, self.on_preview_in_toplevel)
        
    def on_tree_item_delete(self, event):
        selection = self.treeview.selection()
        if selection:
            do_delete = messagebox.askokcancel(
                    _('Delete items'),
                    _('Delete selected items?'),
                    parent=self.treeview.winfo_toplevel())
            
            if do_delete:
                self.on_treeview_delete_selection(None)
    
    def _on_gridrc_changed(self, event):
        # update siblings items that have same row col position
        current_item = self.current_edit
        parent = self.treeview.parent(current_item)
        if parent:
            wmeta = self.treedata[current_item]
            srow = wmeta.layout_property('row')
            scol = wmeta.layout_property('column')
            children = self.treeview.get_children(parent)
            for child in children:
                if child == current_item:
                    continue
                wu = self.treedata[child]
                wu_row = wu.layout_property('row')
                wu_col = wu.layout_property('column')
                if wu_row == srow:
                    wu.copy_gridrc(wmeta, 'row')
                if wu_col == scol:
                    wu.copy_gridrc(wmeta, 'col')
    
    def change_container_manager(self, new_manager):
        item = self.current_edit
        parent = self.treeview.parent(item)
        gridrow = 0
        if parent:
            children = self.treeview.get_children(parent)
            # Stop listening object updates
            self._listen_object_updates = False

            for child in children:
                widget = self.treedata[child]
                widget.manager = new_manager  # Change manager
                
                #update Tree R/C columns
                values = self.treeview.item(child, 'values')
                if new_manager == 'grid':
                    widget.layout_property('row', str(gridrow))
                    widget.layout_property('column', '0')
                    values = (values[0], gridrow, 0)
                    gridrow += 1
                else:
                    values = (values[0], '', '')
                self.treeview.item(child, values=values)
            self._listen_object_updates = True
            self.editor_edit(item, self.treedata[item])
            self.draw_widget(item)
            self.app.set_changed()

    def _on_preview_item_clicked(self, event):
        wid = self.previewer.selected_widget
        logger.debug('item-selected %s', wid)
        self.select_by_id(wid)
        
    def get_children_manager(self, item, current=None):
        '''Get layout manager for children of item'''
        manager = None
        children = self.treeview.get_children(item)
        for child in children:
            if child != current:
                manager = self.treedata[child].manager
                break
        return manager
            
    
    def editor_edit(self, item, wdescr):
        self.current_edit = item
        manager_options = ['grid', 'pack', 'place']
        
        # Determine allowed manager options
        parent = self.treeview.parent(item)
        if parent:
            cm = self.get_children_manager(parent, item)
            if 'grid' == cm:
                manager_options.remove('pack')
            if 'pack' == cm:
                manager_options.remove('grid')
        logger.debug(manager_options)
        
        self.properties_editor.edit(wdescr)
        self.layout_editor.edit(wdescr, manager_options)
        self.bindings_editor.edit(wdescr)

    def editor_hide_all(self):
        self.properties_editor.hide_all()
        self.layout_editor.hide_all()
        self.bindings_editor.hide_all()

    def config_filter(self):
        def on_filtervar_changed(varname, element, mode):
            self.filter_by(self.filtervar.get())

        self.filtervar.trace('w', on_filtervar_changed)

        def on_filterbtn_click():
            self.filtervar.set('')

        self.filter_btn.configure(command=on_filterbtn_click)

    def config_treeview(self):
        """Sets treeview columns and other params"""
        tree = self.treeview
        tree.bind('<Double-1>', self.on_treeview_double_click)
        tree.bind('<<TreeviewSelect>>', self.on_treeview_select, add='+')

    def get_toplevel_parent(self, treeitem):
        """Returns the top level parent for treeitem."""
        tv = self.treeview
        toplevel_items = tv.get_children()

        item = treeitem
        while not (item in toplevel_items):
            item = tv.parent(item)

        return item

    def draw_widget(self, item):
        """Create a preview of the selected treeview item"""
        
        if item:
            self.filter_remove(remember=True)
            
            selected_id = self.treedata[item].identifier
            item = self.get_toplevel_parent(item)
            widget_id = self.treedata[item].identifier
            wclass = self.treedata[item].classname
            uidef = self.tree_to_uidef(item)
            
            self.previewer.draw(item, widget_id, uidef, wclass)
            self.previewer.show_selected(item, selected_id)
            self.filter_restore()

    def on_preview_in_toplevel(self, event=None):
        tv = self.treeview
        sel = tv.selection()
        if sel:
            self.filter_remove(remember=True)
            item = sel[0]
            item = self.get_toplevel_parent(item)
            widget_id = self.treedata[item].identifier
            uidef = self.tree_to_uidef(item)
            self.previewer.preview_in_toplevel(item, widget_id, uidef)
            self.filter_restore()
        else:
            logger.warning(_('No item selected.'))

    def on_treeview_double_click(self, event):
        tv = self.treeview
        sel = tv.selection()
        # toplevel_items = tv.get_children()
        if sel:
            item = sel[0]
            if tv.parent(item) == '':
                # only redraw if toplevel is double clicked
                self.draw_widget(item)

    def on_treeview_delete_selection(self, event=None):
        """Removes selected items from treeview"""

        tv = self.treeview
        selection = tv.selection()

        # Need to remove filter
        self.filter_remove(remember=True)

        toplevel_items = tv.get_children()
        parents_to_redraw = set()
        final_focus = None
        for item in selection:
            try:
                parent = ''
                if item not in toplevel_items:
                    parent = self.get_toplevel_parent(item)
                else:
                    self.previewer.delete(item)
                # determine final focus
                if final_focus is None:
                    candidates = (tv.prev(item), tv.next(item), tv.parent(item))
                    for c in candidates:
                        if c and (c not in selection):
                            final_focus = c
                            break
                # remove item
                del self.treedata[item]
                tv.delete(item)
                self.app.set_changed()
                if parent and (parent not in selection):
                    parents_to_redraw.add(parent)
                self.editor_hide_all()
            except tk.TclError:
                # Selection of parent and child items ??
                # TODO: notify something here
                pass
        # redraw widgets
        for item in parents_to_redraw:
            self.draw_widget(item)
        # Set final item focused
        if final_focus:
            selected_id = self.treedata[final_focus].identifier
            tv.after_idle(lambda: tv.selection_set(final_focus))
            tv.after_idle(lambda: tv.focus(final_focus))
            tv.after_idle(lambda: tv.see(final_focus))
            tv.after_idle(lambda i=final_focus,s=selected_id: self.previewer.show_selected(i, s))
        
        # restore filter
        self.filter_restore()
        
    def new_uidefinition(self):
        author = 'PygubuDesigner {0}'.format(pygubudesigner.__version__)
        uidef = UIDefinition(wmetaclass=WidgetMeta)
        uidef.author = author
        return uidef

    def tree_to_uidef(self, treeitem=None):
        """Traverses treeview and generates a ElementTree object"""

        # Need to remove filter or hidden items will not be saved.
        self.filter_remove(remember=True)
        
        uidef = self.new_uidefinition()
        if treeitem is None:
            items = self.treeview.get_children()
            for item in items:
                node = self.build_uidefinition(uidef, '', item)
                uidef.add_xmlnode(node)
        else:
            node = self.build_uidefinition(uidef, '', treeitem)
            uidef.add_xmlnode(node)

        # restore filter
        self.filter_restore()

        return uidef

    def build_uidefinition(self, uidef, parent, item):
        """Traverses tree and build ui definition"""

        node = uidef.widget_to_xmlnode(self.treedata[item])

        children = self.treeview.get_children(item)
        for child in children:
            child_node = self.build_uidefinition(uidef, item, child)
            uidef.add_xmlchild(node, child_node)
        return node

    def _insert_item(self, root, data, from_file=False):
        """Insert a item on the treeview and fills columns from data"""

        data.setup_defaults() # load default settings for properties and layout
        tree = self.treeview
        treelabel = '{0}: {1}'.format(data.identifier, data.classname)
        row = col = ''
        if root != '' and data.has_layout_defined():
            if data.manager == 'grid':
                row = data.layout_property('row')
                col = data.layout_property('column')
                # fix row position when using copy and paste
                # If collision, increase by 1
                row_count = self.get_max_row(root)
                if not from_file and (row_count > int(row) and int(col) == 0):
                    row = str(row_count + 1)
                    data.layout_property('row', row)

        image = ''
        try:
            image = StockImage.get('16x16-tk.default')
        except StockImageException:
            # TODO: notify something here
            pass

        try:
            image = StockImage.get('16x16-{0}'.format(data.classname))
        except StockImageException:
            # TODO: notify something here
            pass

        values = (data.classname, row, col)
        item = tree.insert(root, 'end', text=treelabel, values=values,
                           image=image)
        data.attach(self)
        self.treedata[item] = data

        self.app.set_changed()

        return item

    def copy_to_clipboard(self):
        """
        Copies selected items to clipboard.
        """
        tree = self.treeview
        # get the selected item:
        selection = tree.selection()
        logger.debug('Selection %s', selection)
        if selection:
            self.filter_remove(remember=True)
            
            uidef = self.new_uidefinition()
            for item in selection:
                node = self.build_uidefinition(uidef, '', item)
                uidef.add_xmlnode(node)
            text = str(uidef)
            tree.clipboard_clear()
            tree.clipboard_append(text)
            self.filter_restore()

    def cut_to_clipboard(self):
        self.copy_to_clipboard()
        self.on_treeview_delete_selection()

    def _validate_add(self, root_item, classname, show_warnings=True):
        is_valid = True

        new_boclass = CLASS_MAP[classname].builder
        root = root_item
        if root:
            root_classname = self.treedata[root].classname
            root_boclass = CLASS_MAP[root_classname].builder
            allowed_children = root_boclass.allowed_children
            if allowed_children:
                if classname not in allowed_children:
                    if show_warnings:
                        str_children = ', '.join(allowed_children)
                        msg = _('Allowed children: %s.')
                        logger.warning(msg, str_children)
                    is_valid = False
                    return is_valid

            children_count = len(self.treeview.get_children(root))
            maxchildren = root_boclass.maxchildren
            if maxchildren is not None and children_count >= maxchildren:
                if show_warnings:
                    msg = trlog(_('Only {0} children allowed for {1}'),
                                maxchildren, root_classname)
                    logger.warning(msg)
                is_valid = False
                return is_valid

            allowed_parents = new_boclass.allowed_parents
            if (allowed_parents is not None and
               root_classname not in allowed_parents):
                if show_warnings:
                    msg = trlog(_('{0} not allowed as parent of {1}'),
                                root_classname, classname)
                    logger.warning(msg)
                is_valid = False
                return is_valid

            if allowed_children is None and root_boclass.container is False:
                if show_warnings:
                    msg = _('Not allowed, %s is not a container.')
                    logger.warning(msg, root_classname)
                is_valid = False
                return is_valid

        else:
            # allways show warning when inserting in top level
            # if insertion is at top level,
            # Validate if it can be added at root level
            allowed_parents = new_boclass.allowed_parents
            if allowed_parents is not None and 'root' not in allowed_parents:
                if show_warnings:
                    msg = _('%s not allowed at root level')
                    logger.warning(msg, classname)
                is_valid = False
                return is_valid

            # if parents are not specified as parent,
            # check that item to insert is a container.
            # only containers are allowed at root level
            if new_boclass.container is False:
                if show_warnings:
                    msg = _('Not allowed at root level, %s is not a container.')
                    logger.warning(msg, classname)
                is_valid = False
                return is_valid
        return is_valid

    def _is_unique_id(self, root, widget_id):
        unique = True
        if root != '':
            data = self.treedata[root]
            if data.identifier == widget_id:
                unique = False
        if unique is True:
            for item in self.treeview.get_children(root):
                unique = self._is_unique_id(item, widget_id)
                if unique is False:
                    break
        return unique

    def _generate_id(self, classname, index):
        class_base_name = classname.split('.')[-1]
        name = class_base_name
        name = '{0}{1}'.format(name, index)
        if pref.get_option('widget_naming_separator') == 'UNDERSCORE':
            name = '{0}_{1}'.format(name, index)
        name = name.lower()
        if pref.get_option('widget_naming_ufletter') == 'yes':
            name = name.capitalize()
        return name

    def get_unique_id(self, classname, start_id=None):
        if start_id is None:
            self.counter[classname] += 1
            start_id = self._generate_id(classname, self.counter[classname])

        is_unique = self._is_unique_id('', start_id)
        while is_unique is False:
            self.counter[classname] += 1
            start_id = self._generate_id(classname, self.counter[classname])
            is_unique = self._is_unique_id('', start_id)
        
        return start_id

    def paste_from_clipboard(self):
        self.filter_remove(remember=True)
        
        tree = self.treeview
        selected_item = ''
        selection = tree.selection()
        if selection:
            selected_item = selection[0]
        try:
            text = tree.selection_get(selection='CLIPBOARD')
            
            uidef = self.new_uidefinition()
            uidef.load_from_string(text)
            for wmeta in uidef.widgets():
                if self._validate_add(selected_item, wmeta.classname):
                    self.update_layout(selected_item, wmeta)
                    self.populate_tree(selected_item, uidef, wmeta)
        except ET.ParseError:
            msg = 'The clipboard does not have a valid widget xml definition.'
            logger.error(msg)
        except tk.TclError:
            pass
        
        if selected_item == '':
            # redraw all
            children = tree.get_children('')
            for child in children:
                self.draw_widget(child)
        else:
            self.draw_widget(selected_item)
            
        
        self.filter_restore()

    def update_layout(self, root, data):
        '''Removes layout info from element, when copied from clipboard.'''
        
        cmanager = self.get_children_manager(root)
        cmanager = cmanager if cmanager is not None else self.default_layout_manager
        emanager = data.manager
        if emanager != 'place' and cmanager != emanager:
            data.manager = cmanager
    
    def add_widget(self, wclass):
        """Adds a new item to the treeview."""

        tree = self.treeview
        #  get the selected item:
        selected_item = ''
        tsel = tree.selection()
        if tsel:
            selected_item = tsel[0]

        #  Need to remove filter if set
        self.filter_remove()

        root = selected_item
        #  check if the widget can be added at selected point
        parent = tree.parent(root)
        has_parent = (parent != root)
        show_warnings = False if has_parent else True
        if not self._validate_add(root, wclass, show_warnings):
            #  if not try to add at item parent level
            parent = tree.parent(root)
            if parent != root:
                logger.info('Failed to add widget, trying one level up.')
                if self._validate_add(parent, wclass):
                    root = parent
                else:
                    return
            else:
                return

        #  root item should be set at this point
        #  setup properties
        parent = None
        if root:
            parent = self.treedata[root]
        manager = self.default_layout_manager # << DEFAULT LAYOUT MANAGER
        if parent is not None:
            cmanager = self.get_children_manager(root)
            manager = cmanager if cmanager else manager
            
        widget_id = self.get_unique_id(wclass)
        pdefaults, ldefaults = WidgetMeta.get_widget_defaults(wclass, widget_id)
        data = WidgetMeta(wclass, widget_id, manager, pdefaults, ldefaults)
        
        # Recalculate position if manager is grid
        if manager == 'grid':
            rownum = '0'
            if root:
                rownum = str(self.get_max_row(root)+1)
            data.layout_property('row', rownum)
            data.layout_property('column', '0')

        item = self._insert_item(root, data)

        # Do redraw
        self.draw_widget(item)

        # Select and show the item created
        tree.after_idle(lambda: tree.selection_set(item))
        tree.after_idle(lambda: tree.focus(item))
        tree.after_idle(lambda: tree.see(item))

    def remove_all(self):
        self.treedata = {}
        self.filter_remove()
        children = self.treeview.get_children()
        if children:
            self.treeview.delete(*children)
        self.editor_hide_all()
        self.counter.clear() # Reset the widget counter (August 19, 2021)

    def load_file(self, filename):
        """Load file into treeview"""

        self.counter.clear()      
        uidef = UIDefinition(wmetaclass=WidgetMeta)
        uidef.load_file(filename)

        self.remove_all()
        self.previewer.remove_all()
        self.editor_hide_all()

        dirname = os.path.dirname(os.path.abspath(filename))
        self.previewer.resource_paths.append(dirname)
        for widget in uidef.widgets():
            self.populate_tree('', uidef, widget,from_file=True)
        
        children = self.treeview.get_children('')
        for child in children:
            self.draw_widget(child)
        self.previewer.show_selected(None, None)

    def populate_tree(self, master, uidef, wmeta, from_file=False):
        """Reads xml nodes and populates tree item"""

        cname = wmeta.classname
        original_id = wmeta.identifier
        uniqueid = self.get_unique_id(cname, wmeta.identifier)
        wmeta.widget_property('id', uniqueid)

        if cname in CLASS_MAP:
            pwidget = self._insert_item(master, wmeta, from_file=from_file)
            for mchild in uidef.widget_children(original_id):
                self.populate_tree(pwidget, uidef, mchild, from_file=from_file)
        else:
            raise Exception('Class "{0}" not mapped'.format(cname))

    def get_max_row(self, item):
        tree = self.treeview
        max_row = -1
        children = tree.get_children(item)
        for child in children:
            row = self.treedata[child].layout_property('row')
            row = int(row)
            if row > max_row:
                max_row = row
        return max_row

    def on_treeview_select(self, event):
        tree = self.treeview
        sel = tree.selection()
        if sel:
            item = sel[0]
            top = self.get_toplevel_parent(item)
            selected_id = self.treedata[item].identifier
            self.previewer.show_selected(top, selected_id)
            # max_rc = self.get_max_row_col(item)
            self.editor_edit(item, self.treedata[item])
        else:
            # No selection hide all
            self.editor_hide_all()

    def get_max_row_col(self, item):
        tree = self.treeview
        max_row = 0
        max_col = 0
        children = tree.get_children(item)
        for child in children:
            data = self.treedata[child]
            row = int(data.layout_property('row'))
            col = int(data.layout_property('column'))
            if row > max_row:
                max_row = row
            if col > max_col:
                max_col = col
        return (max_row, max_col)

    def update_event(self, hint, obj):
        """Updates tree colums when itemdata is changed."""
        
        if not self._listen_object_updates:
            return
        
        tree = self.treeview
        data = obj
        item = self.get_item_by_data(obj)
        item_text = '{0}: {1}'.format(data.identifier, data.classname)
        if item:
            if item_text != tree.item(item, 'text'):
                tree.item(item, text=item_text)
            # if tree.parent(item) != '' and 'layout' in data:
            if tree.parent(item) != '':
                if data.manager == 'grid':
                    row = data.layout_property('row')
                    col = data.layout_property('column')
                    values = tree.item(item, 'values')
                    if (row != values[1] or col != values[2]):
                        values = (data.classname, row, col)
                    tree.item(item, values=values)
            self.draw_widget(item)
            self.app.set_changed()

    def get_item_by_data(self, data):
        skey = None
        for key, value in self.treedata.items():
            if value == data:
                skey = key
                break
        return skey
    
    def on_item_nav_up(self, event=None):
        '''Move selection to prev item'''
        tree = self.treeview
        sel = tree.selection()
        if sel:
            item = sel[0]
            prev = tree.prev(item)
            if prev and tree.item(prev, 'open'):
                children = tree.get_children(prev)
                if children:
                    prev = children[-1]
            if not prev:
                prev = tree.parent(item)
            if prev:
                tree.selection_set(prev)
    
    def on_item_nav_down(self, event=None):
        '''Move selection to next item'''
        tree = self.treeview
        sel = tree.selection()
        if sel:
            item = sel[0]
            next_ = None
            if tree.item(item, 'open'):
                # children
                children = tree.get_children(item)
                if children:
                    next_ = children[0]
            if not next_:
                # sibling
                next_ = tree.next(item)
            if not next_:
                # parent sibling
                parent = tree.parent(item)
                next_ = tree.next(parent)
            if next_:
                tree.selection_set(next_)
    
    def on_item_move_up(self, event):
        tree = self.treeview
        sel = tree.selection()
        if sel:
            self.filter_remove(remember=True)
            item = sel[0]
            parent = tree.parent(item)
            prev = tree.prev(item)
            if prev:
                prev_idx = tree.index(prev)
                tree.move(item, parent, prev_idx)
                manager = self.treedata[item].manager
                self.app.set_changed()
                if manager in ('pack', 'place'):
                    self.draw_widget(item)
            self.filter_restore()

    def on_item_move_down(self, event):
        tree = self.treeview
        sel = tree.selection()
        if sel:
            self.filter_remove(remember=True)
            item = sel[0]
            parent = tree.parent(item)
            next = tree.next(item)
            if next:
                next_idx = tree.index(next)
                tree.move(item, parent, next_idx)
                manager = self.treedata[item].manager
                self.app.set_changed()
                if manager in ('pack', 'place'):
                    self.draw_widget(item)
            self.filter_restore()

    #
    # Item grid move functions
    #
    def on_item_grid_move(self, direction):
        tree = self.treeview
        selection = tree.selection()
        if selection:
            item_first = selection[0]
            self.filter_remove(remember=True)
            for item in selection:
                data = self.treedata[item]
                
                if data.manager != 'grid':
                    break

                if direction == self.GRID_UP:
                    row = int(data.layout_property('row'))
                    if row > 0:
                        row = row - 1
                        data.layout_property('row', str(row))
                        data.notify()
                elif direction == self.GRID_DOWN:
                    row = int(data.layout_property('row'))
                    row = row + 1
                    data.layout_property('row', str(row))
                    data.notify()
                elif direction == self.GRID_LEFT:
                    column = int(data.layout_property('column'))
                    if column > 0:
                        column = column - 1
                        data.layout_property('column', str(column))
                        data.notify()
                elif direction == self.GRID_RIGHT:
                    column = int(data.layout_property('column'))
                    column = column + 1
                    data.layout_property('column', str(column))
                    data.notify()
                root = tree.parent(item)
            self.filter_restore()
            self.editor_edit(item_first, self.treedata[item_first])

    #
    # Filter functions
    #
    def filter_by(self, string):
        """Filters treeview"""

        self._reatach()
        if string == '':
            self.filter_remove()
            return

        self._expand_all()
        self.treeview.selection_set('')

        children = self.treeview.get_children('')
        for item in children:
            _, detached = self._detach(item)
            if detached:
                self._detached.extend(detached)
        for i, p, idx in self._detached:
            # txt = self.treeview.item(i, 'text')
            self.treeview.detach(i)
        self.filter_on = True

    def filter_remove(self, remember=False):
        if self.filter_on:
            sitem = None
            selection = self.treeview.selection()
            if selection:
                sitem = selection[0]
                self.treeview.after_idle(lambda: self._see(sitem))
            if remember:
                self.filter_prev_value = self.filtervar.get()
                self.filter_prev_sitem = sitem
            self._reatach()
            self.filtervar.set('')
        self.filter_on = False

    def filter_restore(self):
        if self.filter_prev_value:
            self.filtervar.set(self.filter_prev_value)
            item = self.filter_prev_sitem
            if item and self.treeview.exists(item):
                self.treeview.selection_set(item)
                self.treeview.after_idle(lambda: self._see(item))
            # clear
            self.filter_prev_value = ''
            self.filter_prev_sitem = None

    def _see(self, item):
        # The item may have been deleted.
        try:
            self.treeview.see(item)
        except tk.TclError:
            pass

    def _expand_all(self, rootitem=''):
        children = self.treeview.get_children(rootitem)
        for item in children:
            self._expand_all(item)
        if rootitem != '' and children:
            self.treeview.item(rootitem, open=True)

    def _reatach(self):
        """Reinsert the hidden items."""
        for item, p, idx in self._detached:
            # The item may have been deleted.
            if self.treeview.exists(item) and self.treeview.exists(p):
                self.treeview.move(item, p, idx)
        self._detached = []

    def _detach(self, item):
        """Hide items from treeview that do not match the search string."""
        to_detach = []
        children_det = []
        children_match = False
        match_found = False

        value = self.filtervar.get()
        txt = self.treeview.item(item, 'text').lower()
        if value in txt:
            match_found = True
        else:
            class_txt = self.treedata[item].classname.lower()
            if value in class_txt:
                match_found = True

        parent = self.treeview.parent(item)
        idx = self.treeview.index(item)
        children = self.treeview.get_children(item)
        if children:
            for child in children:
                match, detach = self._detach(child)
                children_match = children_match | match
                if detach:
                    children_det.extend(detach)

        if match_found:
            if children_det:
                to_detach.extend(children_det)
        else:
            if children_match:
                if children_det:
                    to_detach.extend(children_det)
            else:
                to_detach.append((item, parent, idx))
        match_found = match_found | children_match
        return match_found, to_detach

    #
    # End Filter functions
    #
    
    def get_topwidget_list(self):
        wlist = []
        children = self.treeview.get_children('')
        for item in children:
            data = self.treedata[item]
            label = u'{0} ({1})'.format(data.identifier, data.classname)
            element = (item, label)
            wlist.append(element)
        return wlist
    
    def get_widget_class(self, item):
        return self.treedata[item].classname
    
    def get_widget_id(self, item):
        return self.treedata[item].identifier
    
    def select_by_id(self, widget_id):
        found = None
        for item, data in self.treedata.items():
            if widget_id == data.identifier:
                found = item
                break
        if found:
            tree = self.treeview
            self.filter_remove()
            self._expand_all()
            tree.after_idle(lambda: tree.selection_set(found))
            tree.after_idle(lambda: tree.focus(found))
            tree.after_idle(lambda: tree.see(found))
    
    def is_id_unique(self, idvalue):
        "Check if idvalue is unique in all UI tree."
        # Used in ID validation
        return self._is_unique_id('', idvalue)
    




import os
import pygubu
from PIL import Image, ImageTk


PROJECT_PATH = os.path.dirname(__file__)
PROJECT_UI = os.path.join(PROJECT_PATH, "demoapp.ui")


class DemoApp:
    def __init__(self):
        self.builder = builder = pygubu.Builder()
        builder.add_resource_path(PROJECT_PATH)
        builder.add_from_file(PROJECT_UI)
        self.mainwindow = builder.get_object('topdemo')
        builder.connect_callbacks(self)
        
        canvas = builder.get_object('canvas1')
        # Load image in canvas
        fpath = os.path.join(PROJECT_PATH, 'seaside400.jpg')
        aux = Image.open(fpath)
        self.img = ImageTk.PhotoImage(aux)
        canvas.create_image(0, 0, image=self.img, anchor='nw')
        
    def run(self):
        self.mainwindow.mainloop()

if __name__ == '__main__':
    app = DemoApp()
    app.run()




from __future__ import unicode_literals
import re
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from .propertyeditor import PropertyEditor, register_editor


class RelativeEntryPropertyEditor(PropertyEditor):
    def _create_ui(self):
        self._scheduled = None
        self.scalevar = tk.DoubleVar()
        self.entry = entry = ttk.Entry(self, textvariable=self._variable)
        self.entry.configure(width='4')
        self.entry.pack(side='left')
        
        self.scale = ttk.Scale(self, variable=self.scalevar,
                               command=self.on_scale_changed)
        self.scale.configure(from_='0', orient='horizontal', to='1', value='0')
        self.scale.pack(expand='true', fill='both', side='left')
        
        entry.bind('<FocusOut>', self._on_variable_changed)
        entry.bind('<KeyPress>', self._on_keypress)
    
    def _validate(self):
        is_valid = False
        value = self._get_value()
        vlen = len(value)
        if vlen == 0:
            is_valid = True  # Allow empty string
        else:
            try:
                float(value)
                is_valid = True
            except ValueError:
                pass
        self.show_invalid(not is_valid)
        return is_valid
    
    def on_scale_changed(self, value):
        value = round(self.scalevar.get(), 2)
        self._variable.set(value)
        if self._scheduled is None:
            self._scheduled = self.scale.after(1000, self._update_after)
    
    def _update_after(self):
        self._on_variable_changed()
        self._scheduled = None
    
    def _set_value(self, value):
        """Save value on storage"""
        self._variable.set(value)
        # if empty value, set scale value to zero
        scale_value = 0 if len(value) == 0 else value
        self.scalevar.set(scale_value)
    
    def _after_change(self):
        self.scalevar.set(self._variable.get())


register_editor('relativeentry', RelativeEntryPropertyEditor)


if __name__ == '__main__':
    root = tk.Tk()
    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)

    def make_on_change_cb(editor):
        def on_change_cb(event=None):
            print('Property changed: ')
            print(editor.value)
            print(repr(editor.value))
        return on_change_cb

    editor = RelativeEntryPropertyEditor(root)
    editor.pack(expand=True, fill='x')
    editor.edit('0.5')
    editor.bind('<<PropertyChanged>>', make_on_change_cb(editor))
    root.mainloop()



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here
from __future__ import unicode_literals
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from pygubudesigner.widgets.propertyeditor import *
from pygubudesigner.util.selecttool import SelectTool


class StickyPropertyEditor(PropertyEditor):
    DIM = 3

    def _create_ui(self):
        self._label_var = v = tk.StringVar()
        self._label = w = ttk.Label(self, textvariable=v, width=4)
        w.grid(row=0, column=1)
        
        self._map = {
            '' : [0 , 0, 0, 0 , 1, 0, 0 , 0, 0],

            'n' : [0 , 1, 0, 0 , 0, 0, 0 , 0, 0],
            's' : [0 , 0, 0, 0 , 0, 0, 0 , 1, 0],
            'e' : [0 , 0, 0, 0 , 0, 1, 0 , 0, 0],
            'w' : [0 , 0, 0, 1 , 0, 0, 0 , 0, 0],
            
            'nw' : [1 , 0, 0, 0 , 0, 0, 0 , 0, 0],
            'ne' : [0 , 0, 1, 0 , 0, 0, 0 , 0, 0],
            'sw' : [0 , 0, 0, 0 , 0, 0, 1 , 0, 0],
            'se' : [0 , 0, 0, 0 , 0, 0, 0 , 0, 1],
            
            'new': [1 , 1, 1, 0 , 0, 0, 0 , 0, 0],
            'ew': [0 , 0, 0, 1 , 1, 1, 0 , 0, 0],
            'sew': [0 , 0, 0, 0 , 0, 0, 1 , 1, 1],
            'nsw' : [1 , 0, 0, 1 , 0, 0, 1 , 0, 0],
            'ns' : [0 , 1, 0, 0 , 1, 0, 0 , 1, 0],
            'nse' : [0 , 0, 1, 0 , 0, 1, 0 , 0, 1],
            'nsew' : [1 , 1, 1, 1 , 1, 1, 1 , 1, 1],
            }
        self._items = {}
        self._c = w = tk.Canvas(self, bg='#ffffff', borderwidth=0,
                                highlightthickness=0, width=50, height=50)
        w.bind('<Configure>', self._on_canvas_configure)
        w.bind('<<RegionSelected>>', self._on_region_selected)
        w.grid(row=0, column=0, sticky='w')

        self._tool = t = SelectTool(w)
        
        w.bind('<Button-1>', t.click_handler)
        w.bind('<ButtonRelease-1>', t.release_handler)
        w.bind('<Motion>', t.motion_handler)

        self._on_canvas_configure()
    
    def _set_value(self, value):
        self._variable.set(value)
        self._label_var.set(value)
        self._set_state(value)
    
    def _set_state(self, state):
        self._clear()
        if state in self._map:
            self._paint_state(self._map[state])
        else:
            self._paint_state(self._map[''])

    def _on_canvas_configure(self, event=None):
        # Draw grid
        canvas = self._c
        ch, cw = canvas.winfo_height()-1, canvas.winfo_width()-1
        rw = cw / self.DIM
        rh = ch / self.DIM
        # k = f * DIM + c
        if not self._items:
            for k in range(0, self.DIM * self.DIM):
                f = k % self.DIM
                c = k // self.DIM
                x0 = f * rw
                y0 = c * rh
                x1 = x0 + rw
                y1 = y0 + rh
                item = canvas.create_rectangle(x0, y0, x1, y1, fill='white',
                                               outline='gray')
                self._items[item] = k
        else:
            for item in self._items:
                k = self._items[item]
                f = k % self.DIM
                c = k // self.DIM
                x0 = f * rw
                y0 = c * rh
                x1 = x0 + rw
                y1 = y0 + rh
                canvas.coords(item, x0, y0, x1, y1)
    
    def _on_region_selected(self, event=None):
        canvas = self._c
        region = canvas.region_selected
        items = canvas.find_overlapping(*region)
        self._activate(items)
        
    def _clear(self):
        self._paint(self._items)
    
    def _paint(self, items, color='white'):
        for item in items:
            self._c.itemconfigure(item, fill=color)
    
    def _paint_state(self, state, color='#9999cc'):
        for item in self._items:
            pos = self._items[item]
            if state[pos]:
                self._c.itemconfigure(item, fill=color)
    
    def _activate(self, items):
        self._paint(self._items.keys())
        state = [0] * (self.DIM * self.DIM)
        for item in items:
            pos = self._items[item]
            state[pos] = 1
        # check valid state
        found = ''
        for k, v in self._map.items():
            if v == state:
                found = k
                break
        self._paint_state(self._map[found], color='#9999cc')
        self._set_value(found)
        self._on_variable_changed()


register_editor('stickyentry', StickyPropertyEditor)

if __name__ == '__main__':
    root = tk.Tk()
    editor = StickyPropertyEditor(root)
    editor.grid()
    editor.edit('nsew')

    def see_var(event=None):
        print(editor.value)

    editor.bind('<<PropertyChanged>>', see_var)
    root.mainloop()



# encoding: UTF-8
#
# Copyright 2012-2021 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
import keyword
import re
import json

try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from pygubu.builder.builderobject import CB_TYPES
from pygubudesigner.i18n import translator as _
from pygubudesigner.widgets.propertyeditor import *


class CommandPropertyBase(PropertyEditor):
    RE_IDENTIFIER = re.compile('[_A-Za-z][_a-zA-Z0-9]*$')
    
    def is_safe_identifier(self, value):
        is_valid = True
        if keyword.iskeyword(value):
            is_valid = False
        if is_valid and not self.RE_IDENTIFIER.match(value):
            is_valid = False
        return is_valid


class SimpleCommandEntry(CommandPropertyBase):
    cmd_type = CB_TYPES.SIMPLE
    
    def _create_ui(self):
        self._cbname = w = EntryPropertyEditor(self)
        w.grid(row=0, column=0, sticky='nswe', columnspan=2)
        w.bind('<<PropertyChanged>>', self._on_variable_changed)
        self.columnconfigure(0, weight=1)
    
    def _set_value(self, value):
        """Save value on storage"""
        cbname = ''
        if len(value) != 0:
            vd = json.loads(value)
            cbname = vd['value']
        self._cbname.edit(cbname)
        self._variable.set(value)
    
    def _get_value(self):
        value = ''
        if len(self._cbname.value) != 0:
            cmd = {
                'type': 'command',
                'value': self._cbname.value,
                'cbtype': self.cmd_type
            }
            value = json.dumps(cmd)
        return value
        
    def _validate(self):
        is_valid = True
        value = self._cbname.value
        if len(value) != 0:
            is_valid = self.is_safe_identifier(value)
        self.show_invalid(not is_valid)
        return is_valid


class ScrollCommandEntry(SimpleCommandEntry):
    cmd_type = CB_TYPES.SCROLL


class ScrollSetCommandEntry(SimpleCommandEntry):
    cmd_type = CB_TYPES.SCROLLSET


class ScaleCommandEntry(SimpleCommandEntry):
    cmd_type = CB_TYPES.SCALE


register_editor('simplecommandentry', SimpleCommandEntry)
register_editor('scrollcommandentry', ScrollCommandEntry)
register_editor('scrollsetcommandentry', ScrollSetCommandEntry)
register_editor('scalecommandentry', ScaleCommandEntry)


class CommandPropertyEditor(CommandPropertyBase):
    
    def _create_ui(self):
        self._lbl_callback = _('Callback:')
        self._plabel = w = ttk.Label(self, text=self._lbl_callback,
                                     font='TkSmallCaptionFont')
        w.grid(row=0, column=0, sticky='nswe')
        
        self._cbname = w = EntryPropertyEditor(self)
        w.grid(row=0, column=1, sticky='nswe')
        w.bind('<<PropertyChanged>>', self._on_variable_changed)
        # type label:
        w = ttk.Label(self, text='Type:', font='TkSmallCaptionFont')
        w.grid(row=1, column=0, sticky='nsew')
        # type combo
        cmdtypes = (
            (CB_TYPES.SIMPLE, _('Simple')),
            (CB_TYPES.WITH_WID, _('Send widget ID')),
        )
        self._cmdtype = w = ChoiceByKeyPropertyEditor(self)
        w.parameters(values=cmdtypes)
        w.bind('<<PropertyChanged>>', self._on_variable_changed)
        w.grid(row=1, column=1, sticky='nswe')
        self.columnconfigure(1, weight=1)
    
    def _set_value(self, value):
        """Save value on storage"""
        cbname = ''
        cbtype = CB_TYPES.SIMPLE
        if len(value) != 0:
            vd = json.loads(value)
            cbname = vd['value']
            cbtype = vd['cbtype']
        self._cbname.edit(cbname)
        self._cmdtype.edit(cbtype)
        self._variable.set(value)
    
    def _get_value(self):
        value = ''
        if len(self._cbname.value) != 0:
            cmd = {
                'type': 'command',
                'value': self._cbname.value,
                'cbtype': self._cmdtype.value
            }
            value = json.dumps(cmd)
        return value
        
    def _validate(self):
        is_valid = True
        value = self._cbname.value
        if len(value) != 0:
            is_valid = self.is_safe_identifier(value)
        self.show_invalid(not is_valid)
        return is_valid


register_editor('commandentry', CommandPropertyEditor)



if __name__ == '__main__':
    root = tk.Tk()
    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)
    editor = CommandPropertyEditor(root)
    editor.grid(sticky='nsew')
    
    def change_cb(self, event=None):
        print('change cb')
    
    editor.bind('<<PropertyChanged>>', change_cb)
    value = '{"value":"mycb", "cbtype": "simple"}'
    editor.edit(value)

    root.mainloop()


# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here


class Observable(object):

    def __init__(self):
        super(Observable, self).__init__()
        self._observers = []

    def attach(self, observer):
        if not observer in self._observers:
            self._observers.append(observer)

    def detach(self, observer):
        try:
            self._observers.remove(observer)
        except ValueError:
            pass

    def notify(self, hint=None, modifier=None):
        for observer in self._observers:
            if modifier != observer:
                observer.update_event(hint, self)



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here

try:
    import tkinter as tk
except:
    import Tkinter as tk

from pygubu.builder.builderobject import *
from pygubu.builder.tkstdwidgets import TKToplevel


class ToplevelFramePreview(tk.Frame):

    def __init__(self, master=None, **kw):
        tk.Frame.__init__(self, master, **kw)
        self.tl_attrs = {}
        self._w_set = False
        self._h_set = False

    def configure(self, cnf=None, **kw):
        if kw:
            cnf = tk._cnfmerge((cnf, kw))
        elif cnf:
            cnf = tk._cnfmerge(cnf)
        key = 'width'
        if key in cnf:
            value = int(cnf[key])
            minsize = self.tl_attrs.get('minsize', None)
            maxsize = self.tl_attrs.get('maxsize', None)
#            print(value, minsize, maxsize)
            remove = False
#            print('tl_attrs:', self.tl_attrs)
            if minsize and value < minsize[0]:
                remove = True
            if maxsize and value > maxsize[0]:
                remove = True
            if self._w_set:
                resizable = self.tl_attrs.get('resizable', None)
                if resizable and not TKToplevel.RESIZABLE[resizable][0]:
                    remove = True
            if remove:
#                print('rm', key, value)
                cnf.pop(key)
            else:
                self._w_set = True
        key = 'height'
        if key in cnf:
            value = int(cnf[key])
            minsize = self.tl_attrs.get('minsize', None)
            maxsize = self.tl_attrs.get('maxsize', None)
#            print(value, minsize, maxsize)
            remove = False
            if minsize and value < minsize[1]:
                remove = True
            if maxsize and value > maxsize[1]:
                remove = True
            if self._h_set:
                resizable = self.tl_attrs.get('resizable', None)
                if resizable and not TKToplevel.RESIZABLE[resizable][1]:
                    remove = True
            if remove:
#                print('rm', key, value)
                cnf.pop(key)
            else:
                self._h_set = True
        key = 'menu'
        if key in cnf:
            # No menu preview available
            cnf.pop(key)
        
        return tk.Frame.configure(self, cnf)


class ToplevelFramePreviewBO(BuilderObject):
    class_ = ToplevelFramePreview
    container = True
    #Add fake 'modal' property for Dialog preview
    properties = TKToplevel.properties + ('modal',)
    ro_properties = TKToplevel.ro_properties

    def _set_property(self, target_widget, pname, value):
        tw = target_widget
        tw.tl_attrs[pname] = value
        method_props = ('iconbitmap', 'iconphoto', 'overrideredirect', 'title')
        if pname in method_props:
            pass
        elif pname in ('maxsize', 'minsize'):
            if not value:
                del tw.tl_attrs[pname]
            elif '|' in value:
                w, h = value.split('|')
                if w and h:
                    tw.tl_attrs[pname] = (int(w), int(h))
                else:
                    del tw.tl_attrs[pname]
        elif pname == 'geometry':
            if value:
                dim = value.split('+')[0]
                dim = dim.split('-')[0]
                w, h = dim.split('x')
                if w and h:
                    tw.tl_attrs['minsize'] = (int(w), int(h))
                    tw._h_set = tw._w_set = False
                    tw.configure(width=w, height=h)
                    if tw.pack_slaves():
                        tw.pack_propagate(0)
                    elif tw.grid_slaves():
                        tw.grid_propagate(0)
        elif pname == 'resizable':
            if value:
                if value in ('both', 'horizontally'):
                    tw.columnconfigure(0, weight=1)
                if value in ('both', 'vertically'):
                    tw.rowconfigure(0, weight=1)
        elif pname == 'modal':
            # Do nothing, fake 'modal' property for dialog preview
            pass
        else:
            super(ToplevelFramePreviewBO, self)._set_property(tw, pname, value)


register_widget('pygubudesigner.ToplevelFramePreview',
                ToplevelFramePreviewBO, 'ToplevelFramePreview', tuple())



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here

from __future__ import unicode_literals
from collections import OrderedDict
import sys
import xml.etree.ElementTree as ET
import re
import logging

try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

import pygubu
import pygubudesigner.actions as actions
from pygubu.stockimage import StockImage
from .widgetdescr import WidgetMeta
from .widgets.toplevelframe import ToplevelFramePreview

try:
    basestring
except NameError:
    basestring = str

logger = logging.getLogger(__name__)
RE_FONT = re.compile("(?P<family>\{\w+(\w|\s)*\}|\w+)\s?(?P<size>-?\d+)?\s?(?P<modifiers>\{\w+(\w|\s)*\}|\w+)?")


class BuilderForPreview(pygubu.Builder):
    normalwidgets = ['tk.Menu', 'tk.PanedWindow', 'tk.PanedWindow.Pane',
                     'ttk.Panedwindow', 'ttk.Notebook',
                     'ttk.Panedwindow.Pane', 'ttk.Notebook.Tab',
                     'pygubudesigner.ToplevelFramePreview']
    
    def _pre_realize(self, bobject):
        super(BuilderForPreview, self)._pre_realize(bobject)
        wmeta = bobject.wmeta
        cname = wmeta.classname
        #print('In _pre_process_data', cname)
        
        #  Do not resize main window when
        #  Sizegrip is dragged on preview panel.
        if cname == 'ttk.Sizegrip':
            wmeta.properties['class_'] = 'DUMMY_CLASS'
        if cname not in self.normalwidgets:
            if 'class_' in bobject.properties:
                wmeta.properties['class_'] = 'PreviewWidget'
    
    def _post_realize(self, bobject):
        #print('In postProcess', bobject.wmeta.classname)
        cname = bobject.wmeta.classname
        if cname not in self.normalwidgets:
            if cname.startswith('tk.Menuitem'):
                return
            self.make_previewonly(bobject.widget)
    
    def make_previewonly(self, w):
        '''Make widget just display with no functionality.'''
        for child in w.winfo_children():
            self._crop_widget(child)
        self._crop_widget(w)
    
    def _crop_widget(self, w):
        '''Remove standard widget functionality.'''
        wclass = w.winfo_class()
        bindtags = w.bindtags()
        if wclass in bindtags:
            bindtags = list(bindtags)
            bindtags.remove(wclass)
            w.bindtags(bindtags)
    
    def get_widget_id(self, widget):
        wid = None
        for key, o in self.objects.items():
            if o.widget == widget:
                wid = key
                break
        return wid
    
    def show_selected(self, select_id):
        self._show_notebook_tabs(select_id)
    
    def _show_notebook_tabs(self, select_id):
        xpath = ".//object[@class='ttk.Notebook.Tab']"
        xpath = xpath.format(select_id)
        # find all tabs
        tabs = self.uidefinition.root.findall(xpath)
        if tabs is not None:
            for tab in tabs:
                # check if selected_id is inside this tab
                tab_id = tab.get('id')
                xpath = ".//object[@id='{0}']".format(select_id)
                o = tab.find(xpath)
                if o is not None:
                    # selected_id is inside, find the tab child
                    # and select this tab
                    xpath = './child/object[1]'
                    child = tab.find(xpath)
                    child_id = child.get('id')
                    notebook = self.objects[tab_id].widget
                    current_tab = self.objects[child_id].widget
                    notebook.select(current_tab)
                    #print(select_id, ' inside', tab_id, 'child', child_id)
                    


class Preview(object):
    def __init__(self, id_, canvas, x=0, y=0, rpaths=None):
        self.id = 'preview_{0}'.format(id_)
        self.x = x
        self.y = y
        self.w = 10
        self.h = 10
        self.min_w = self.w
        self.min_h = self.h
        self.resizer_h = 10
        self.canvas = canvas
        self.shapes = {}
        self._create_shapes()
        # --------
        self.builder = None
        self.canvas_window = None
        self._resource_paths = rpaths if rpaths is not None else []
        # --
        self.selected_widget = None
        self.root_widget = None

    def width(self):
        return self.w

    def height(self):
        return self.h + self.resizer_h
        
    def _create_builder(self):
        b = BuilderForPreview()
        for p in self._resource_paths:
            b.add_resource_path(p)
        return b

    def _create_shapes(self):
        # Preview box
        c = self.canvas
        x, y, x2, y2 = (-1001, -1000, -1001, -1000)
        s1 = c.create_rectangle(x, y, x2, y2,
                                width=2, outline='blue', tags=self.id)
        s2 = c.create_rectangle(x, y, x2, y2, fill='blue', outline='blue',
                                tags=(self.id, 'resizer'))
        s3 = c.create_text(x, y, text='widget_id', anchor=tk.NW,
                           fill='white', tags=self.id)
        s4 = c.create_window(x, y, anchor=tk.NW, tags=self.id)
        self.shapes = {
            'outline': s1,
            'resizer': s2,
            'text': s3,
            'window': s4
        }
        self.draw()

    def erase(self):
        self.canvas_window.destroy()
        for key, sid in self.shapes.items():
            self.canvas.delete(sid)

    def draw(self):
        c = self.canvas
        x, y, x2, y2 = (self.x, self.y, self.x + self.w, self.y + self.h)
        c.coords(self.shapes['outline'], x, y, x2, y2)
        tbbox = c.bbox(self.shapes['text'])
        tw, th = tbbox[2] - tbbox[0] + 10, tbbox[3] - tbbox[1] + 6
        self.resizer_h = th
        rx2 = self.x + self.w
        ry2 = self.y + self.h + self.resizer_h
        rx = rx2 - tw
        ry = self.y + self.h
        c.coords(self.shapes['resizer'], rx, ry, rx2, ry2)
        tx = rx + 5
        ty = ry + 3
        c.coords(self.shapes['text'], tx, ty)
        c.coords(self.shapes['window'], x, y)

    def move_by(self, dx, dy):
        self.x += dx
        self.y += dy
        self.draw()

    def resize_to(self, w, h):
        self.resize_by(w - self.w, h - self.h)

    def resize_by(self, dw, dh):
        new_w = self.w + dw
        new_h = self.h + dh
        changed = False
        if new_w >= self.min_w:
            self.w = new_w
            changed = True
        if new_h >= self.min_h:
            self.h = new_h
            changed = True
        if changed:
            self.draw()
            self._resize_preview_window()

    def _resize_preview_window(self):
        if self.canvas_window:
            self.canvas_window.configure(width=self.w, height=self.h)

    def update(self, widget_id, uidefinition):
        # delete current preview
        # FIXME maybe do something to update preview without re-creating all ?
        del self.builder
        self.builder = None
        self.canvas.itemconfigure(self.shapes['window'], window='')
        if self.canvas_window:
            self.canvas_window.destroy()

        # Create preview
        canvas_window = ttk.Frame(self.canvas, style='PreviewFrame.TFrame')
        #canvas_window.rowconfigure(0, weight=1)
        #canvas_window.columnconfigure(0, weight=1)

        self.canvas.itemconfigure(self.shapes['text'], text=widget_id)

        self._preview_widget = \
            self.create_preview_widget(canvas_window, widget_id, uidefinition)
        self.root_widget = self._preview_widget

        self.canvas_window = canvas_window
        self.canvas.itemconfigure(self.shapes['window'], window=canvas_window)
        canvas_window.update_idletasks()
        if canvas_window.pack_slaves():
            canvas_window.pack_propagate(0)
        elif canvas_window.grid_slaves():
            canvas_window.grid_propagate(0)
        self.min_w = self._get_wreqwidth()
        self.min_h = self._get_wreqheight()
        self.w = self.min_w * 2
        self.h = self.min_h * 2
        self.resize_to(self.min_w, self.min_h)

    def create_preview_widget(self, parent, widget_id, uidefinition):
        self.builder = self._create_builder()
        self.builder.uidefinition = uidefinition
        widget = self.builder.get_object(widget_id, parent)
        return widget

    def get_widget_by_id(self, widget_id):
        return self.builder.get_object(widget_id)
    
    def show_selected(self, select_id):
        self.builder.show_selected(select_id)

    def create_toplevel(self, widget_id, uidefinition):
        # Create preview
        builder = pygubu.Builder()
        builder.uidefinition = uidefinition
        top = tk.Toplevel(self.canvas)
        top.columnconfigure(0, weight=1)
        top.rowconfigure(0, weight=1)
        builder.get_object(widget_id, top)
        return top

    def _get_wreqwidth(self):
        return self._preview_widget.winfo_reqwidth()

    def _get_wreqheight(self):
        return self._preview_widget.winfo_reqheight()


class DefaultMenuPreview(Preview):

    def create_preview_widget(self, parent, widget_id, uidefinition):
        self.builder = self._create_builder()
        self.builder.uidefinition = uidefinition
        menubutton = ttk.Menubutton(parent, text='Menu preview')
        menubutton.grid()
        widget = self.builder.get_object(widget_id, menubutton)
        menubutton.configure(menu=widget)
        return menubutton

    def create_toplevel(self, widget_id, uidefinition):
        # Create preview
        builder = pygubu.Builder()
        builder.uidefinition = uidefinition
        top = tk.Toplevel(self.canvas)
        top.columnconfigure(0, weight=1)
        top.rowconfigure(0, weight=1)

        menu = builder.get_object(widget_id, top)
        top['menu'] = menu
        return top

    def resize_by(self, dw, hw):
        return


class OnCanvasMenuPreview(Preview):
    fonts = {}

    def __init__(self, id_, canvas, x=0, y=0, rpaths=None):
        super(OnCanvasMenuPreview, self).__init__(id_, canvas, x, y, rpaths)
        self._menu = None
        self._cwidth = 0
        self._cheight = 0

    def _get_wreqwidth(self):
        return self._cwidth

    def _get_wreqheight(self):
        return self._cheight

    def _get_font(self, font):
        fontname = family = 'TkMenuFont'
        size = 12
        modifiers = ''
        tclobject = False

        if font and isinstance(font, basestring):
            fontname = family = font
        elif isinstance(font, tk._tkinter.Tcl_Obj):
            fontname = family = str(font)
            tclobject = True
        elif isinstance(font, tuple):
            fontname = str(font[4])
            tclobject = True
        if tclobject:
            s = RE_FONT.search(fontname)
            if s:
                g = s.groupdict()
                family = g['family'].replace('{', '').replace('}', '')
                size = g['size']
                modifiers = g['modifiers'] if g['modifiers'] else ''
        if fontname not in OnCanvasMenuPreview.fonts:
            weight = 'bold' if 'bold' in modifiers else 'normal'
            slant = 'italic' if 'italic' in modifiers else 'roman'
            underline = '1' if 'underline' in modifiers else '0'
            overstrike = '1' if 'overstrike' in modifiers else '0'
            kw = {'family': family, 'weight': weight, 'slant': slant,
                  'underline': underline, 'overstrike': overstrike}
            if size:
                kw['size'] = size
            OnCanvasMenuPreview.fonts[fontname] = tk.font.Font(**kw)
        return OnCanvasMenuPreview.fonts[fontname]

    def _calculate_menu_wh(self):
        """ Calculate menu widht and height."""
        w = iw = 50
        h = ih = 0
        # menu.index returns None if there are no choices
        index = self._menu.index(tk.END)
        index = index if index is not None else 0
        count = index + 1

        # First calculate using the font paramters of root menu:
        font = self._menu.cget('font')
        font = self._get_font(font)
        for i in range(0, count):
            mtype = self._menu.type(i)
            if mtype == 'tearoff':
                continue
            label = 'default'
            ifont = 'TkMenuFont'
            if mtype != 'separator':
                label = self._menu.entrycget(i, 'label')
                ifont = self._menu.entrycget(i, 'font')
            wpx = font.measure(label)
            hpx = font.metrics('linespace')
            w += wpx
            if hpx > h:
                h = hpx * 2
            # Calculate using font configured for each subitem
            ifont = self._get_font(ifont)
            wpx = ifont.measure(label)
            hpx = ifont.metrics('linespace')
            iw += wpx
            if hpx > ih:
                ih = hpx * 2
        # Then compare 2 sizes and use the greatest
        w = max(w, iw, 100)
        h = max(h, ih, 25)
        self._cwidth = w + int(w * 0.25)
        self._cheight = h + int(h * 0.25)

    def create_preview_widget(self, parent, widget_id, uidefinition):
        container = tk.Frame(parent, container=True, height=50)
        container.pack(fill='both', expand=True)

        self._top = top = tk.Toplevel(parent, use=container.winfo_id())
        top.maxsize(2048, 50)
        top.resizable(width=True, height=False)
        top.update()

        self.builder = self._create_builder()
        self.builder.uidefinition = uidefinition
        self._menu = widget = self.builder.get_object(widget_id, top)
        top.configure(menu=widget)
        self._calculate_menu_wh()
        return parent

    def create_toplevel(self, widget_id, uidefinition):
        # Create preview
        builder = pygubu.Builder()
        builder.uidefinition = uidefinition
        top = tk.Toplevel(self.canvas)
        top.columnconfigure(0, weight=1)
        top.rowconfigure(0, weight=1)

        menu = builder.get_object(widget_id, top)
        top['menu'] = menu
        return top


MenuPreview = DefaultMenuPreview
if sys.platform == 'linux':
    MenuPreview = OnCanvasMenuPreview


class ToplevelPreview(Preview):

    def create_preview_widget(self, parent, widget_id, uidefinition):
        # Change real Toplevel for a preview replacement:
        # Add same behavior of Toplevel. Default expand both sides:
        old = uidefinition.get_widget(widget_id)
        newroot = WidgetMeta('pygubudesigner.ToplevelFramePreview',
                             widget_id, 'pack')
        newroot.copy_properties(old)
        #newroot.widget_property('height', '200')
        #newroot.widget_property('width', '200')
        newroot.layout_property('expand', 'true')
        newroot.layout_property('fill', 'both')
        
        uidefinition.replace_widget(widget_id, newroot)
        # end add behaviour

        self.builder = self._create_builder()
        self.builder.uidefinition = uidefinition
        widget = self.builder.get_object(widget_id, parent)
        return widget

    def create_toplevel(self, widget_id, uidefinition):
        # Create preview
        builder = pygubu.Builder()
        builder.uidefinition = uidefinition
        top = builder.get_object(widget_id, self.canvas)
        return top


class DialogPreview(ToplevelPreview):
    def create_toplevel(self, widget_id, uidefinition):
        top = super(DialogPreview, self).create_toplevel(widget_id,
                                                         uidefinition)
        top.run()
        return top



class PreviewHelper:
    indicators_tag = ('nw', 'ne', 'sw', 'se')

    def __init__(self, canvas):
        self.canvas = canvas
        self.previews = OrderedDict()
        self.padding = 20
        self.indicators = None
        self._sel_id = None
        self._sel_widget = None
        self.toplevel_previews = []
        self.resource_paths = []

        self._moving = False
        self._last_event = None
        self._objects_moving = None
        canvas.bind('<Button-1>', self.click_handler)
        canvas.bind('<Button-1>', lambda e: canvas.focus_set(), add=True)
        canvas.bind('<ButtonRelease-1>', self.release_handler)
        canvas.bind('<Motion>', self.motion_handler)
        canvas.bind('<4>', lambda event: canvas.yview('scroll', -1, 'units'))
        canvas.bind('<5>', lambda event: canvas.yview('scroll', 1, 'units'))
        self._create_indicators()
        
        s = ttk.Style()
        s.configure('PreviewFrame.TFrame', background='lightgreen')
        
        self.selected_widget = None
        canvas.bind_all(actions.PREVIEW_TOPLEVEL_CLOSE_ALL,
                    lambda e: self.close_toplevel_previews())
    
    def add_resource_path(self, path):
        self.resource_paths.append(path)

    def motion_handler(self, event):
        if not self._moving:
            c = event.widget
            x = c.canvasx(event.x)
            y = c.canvasy(event.y)
            if self._over_resizer(x, y):
                c.configure(cursor='fleur')
            else:
                c.configure(cursor='')
        else:
            dx = event.x - self._last_event.x
            dy = event.y - self._last_event.y
            self._last_event = event
            if dx or dy:
                self.resize_preview(dx, dy)

    def click_handler(self, event):
        c = event.widget
        x = c.canvasx(event.x)
        y = c.canvasy(event.y)

        if self._over_resizer(x, y):
            ids = c.find_overlapping(x, y, x, y)
            if ids:
                self._moving = True
                self._objects_moving = ids
                c.configure(cursor='fleur')
                self._last_event = event

    def release_handler(self, event):
        self._objects_moving = None
        self._moving = False

    def _over_resizer(self, x, y):
        "Returns True if mouse is over a resizer"

        over_resizer = False
        c = self.canvas
        ids = c.find_overlapping(x, y, x, y)
        if ids:
            o = ids[0]
            tags = c.gettags(o)
            if 'resizer' in tags:
                over_resizer = True
        return over_resizer

    def resize_preview(self, dw, dh):
        "Resizes preview that is currently dragged"

        # identify preview
        if self._objects_moving:
            id_ = self._objects_moving[0]
            tags = self.canvas.gettags(id_)
            for tag in tags:
                if tag.startswith('preview_'):
                    _, ident = tag.split('preview_')
                    preview = self.previews[ident]
                    preview.resize_by(dw, dh)
                    self.move_previews()
                    break
        self._update_cregion()

    def _update_cregion(self):
        # update canvas scrollregion
        bbox = self.canvas.bbox(tk.ALL)
        padd = 20
        if bbox is not None:
            region = (0, 0, bbox[2] + padd, bbox[3] + padd)
            self.canvas.configure(scrollregion=region)

    def move_previews(self):
        "Move previews after a resize event"

        # calculate new positions
        min_y = self._calc_preview_ypos()
        for idx, (key, p) in enumerate(self.previews.items()):
            new_dy = min_y[idx] - p.y
            self.previews[key].move_by(0, new_dy)
        self._update_cregion()
        self.show_selected(self._sel_id, self._sel_widget)

    def _calc_preview_ypos(self):
        "Calculates the previews positions on canvas"

        y = 10
        min_y = [y]
        for k, p in self.previews.items():
            y += p.height() + self.padding
            min_y.append(y)
        return min_y

    def _get_slot(self):
        "Returns the next coordinates for a preview"

        x = y = 10
        for k, p in self.previews.items():
            y += p.height() + self.padding
        return x, y

    def draw(self, identifier, widget_id, uidefinition, wclass):
        logger.debug('Preview.draw: %s,%s', identifier, widget_id)
        preview_class = Preview
        if wclass == 'tk.Menu':
            preview_class = MenuPreview
        elif wclass == 'tk.Toplevel':
            preview_class = ToplevelPreview
        elif wclass == 'pygubu.builder.widgets.dialog':
            preview_class = DialogPreview
        if identifier not in self.previews:
            x, y = self._get_slot()
            self.previews[identifier] = preview \
                = preview_class(identifier, self.canvas, x, y,
                                self.resource_paths)
        else:
            preview = self.previews[identifier]
        preview.update(widget_id, uidefinition)
        self.reset_selected(identifier)
        self.move_previews()
        
        if True:
            # we are allways recreating full preview.
            # create callback and bind widgets
            def callback(event, self=self, previewid=identifier):
                self.preview_click_handler(previewid, event)
            widget = preview.root_widget
            self.bind_preview_widget(widget, callback)            

    def _create_indicators(self):
        # selected indicators
        self.indicators = []
        anchors = {'nw': tk.SE, 'ne': tk.SW, 'sw': tk.NE, 'se': tk.NW}
        for sufix in self.indicators_tag:
            label = tk.Label(self.canvas,
                             image=StockImage.get('indicator_' + sufix))
            self.indicators.append(label)
            self.canvas.create_window(-10, -10, anchor=anchors[sufix],
                                      window=label, tags=sufix)

    def _calculate_indicator_coords(self, tag, widget):
        x = y = 0
        wx = widget.winfo_rootx()
        wy = widget.winfo_rooty()
        ww = widget.winfo_width()
        wh = widget.winfo_height()
        cx = self.canvas.winfo_rootx()
        cy = self.canvas.winfo_rooty()
        if tag == 'nw':
            x = wx - cx
            y = wy - cy
        if tag == 'ne':
            x = (wx - cx) + ww
            y = (wy - cy)
        if tag == 'sw':
            x = (wx - cx)
            y = (wy - cy) + wh
        if tag == 'se':
            x = (wx - cx) + ww
            y = (wy - cy) + wh
        x, y = self.canvas.canvasx(x), self.canvas.canvasy(y)
        return (x, y)

    def show_selected(self, identifier, selected_id=None):
        canvas = self.canvas
        if selected_id is None:
            for indicator in self.indicators_tag:
                canvas.itemconfigure(indicator, state=tk.HIDDEN)
        elif identifier in self.previews:
            for indicator in self.indicators_tag:
                canvas.itemconfigure(indicator, state=tk.NORMAL)
            preview = self.previews[identifier]
            canvas.update_idletasks()
            preview.show_selected(selected_id)
            widget = preview.get_widget_by_id(selected_id)
            for indicatorw in self.indicators:
                try:
                    indicatorw.lift(widget)
                except tk.TclError:
                    pass
            for tag in self.indicators_tag:
                x, y = self._calculate_indicator_coords(tag, widget)
                ox, oy = canvas.coords(tag)
                canvas.move(tag, x - ox, y - oy)
        self._sel_id = identifier
        self._sel_widget = selected_id

    def delete(self, identifier):
        if identifier in self.previews:
            preview = self.previews[identifier]
            preview.erase()
            del self.previews[identifier]
            self.reset_selected(identifier)
            self.move_previews()

    def reset_selected(self, identifier):
        if identifier == self._sel_id:
            self._sel_id = None
            self._sel_widget = None

    def remove_all(self):
        for identifier in list(self.previews):
            self.delete(identifier)
        self.resource_paths = []

    def preview_in_toplevel(self, identifier, widget_id, uidefinition):
        preview = self.previews[identifier]
        top = preview.create_toplevel(widget_id, uidefinition)
        self.toplevel_previews.append(top)

    def close_toplevel_previews(self):
        for top in self.toplevel_previews:
            top.destroy()
        self.toplevel_previews = []
    
    def preview_click_handler(self, preview_id, event=None):
        preview = self.previews[preview_id]
        wid = preview.builder.get_widget_id(event.widget)
        if wid is not None:
            self.selected_widget = wid
            self.canvas.event_generate('<<PreviewItemSelected>>')
            self.canvas.focus_set() # add focus to canvas
    
    def bind_preview_widget(self, widget, callback):
        widget.bind('<Button-1>', callback)
        for w in widget.winfo_children():
            self.bind_preview_widget(w, callback)



# encoding: utf-8
import os
import sys
import logging
from mako.lookup import TemplateLookup

try:
    import tkinter as tk
    from tkinter import ttk
    from tkinter import filedialog
    from tkinter import messagebox
except:
    import Tkinter as tk
    import ttk
    import tkMessageBox as messagebox
    import tkFileDialog as filedialog
import pygubu
from .codebuilder import UI2Code

logger = logging.getLogger(__name__)
CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMPLATE_DIR = os.path.join(CURRENT_DIR, 'template')
print('TEMPLATE_DIR', TEMPLATE_DIR)
makolookup = TemplateLookup(directories=[TEMPLATE_DIR])

class ScriptGenerator(object):
    def __init__(self, app):
        self.app = app
        self.builder = builder = app.builder
        self.tree = app.tree_editor
        self.projectname = ''

        self.widgetlist = builder.get_object('widgetlist')
        
        self.widgetlistvar = None
        self.widgetlist_keyvar = None
        self.template_var = None
        self.classnamevar = None
        self.template_desc_var = None
        self.import_tkvars_var = None
        myvars = ['widgetlistvar', 'widgetlist_keyvar', 'template_var',
                  'classnamevar', 'template_desc_var', 'import_tkvars_var']
        builder.import_variables(self, myvars)
        
        self.txt_code = builder.get_object('txt_code')
        self.cb_import_tkvars = builder.get_object('cb_import_tkvars')
        
        _ = self.app.translator
        self.msgtitle = _('Script Generator')
        
        self.template_desc = {
            'application': _('Create a pygubu application script using the UI definition.'),
            'codescript': _('Create a coded version of the UI definition.'),
            'widget': _('Create a base class for your custom widget.')
        }
        self.template_var.set('application')
        self.import_tkvars_var.set(True)
        
    def camel_case(self, st):
        output = ''.join(x for x in st.title() if x.isalnum())
        return output
    
    def on_code_generate_clicked(self):
        if self.form_valid():
            tree_item = self.widgetlist_keyvar.get()
            template = self.template_var.get()
            generator = UI2Code()
            uidef = self.tree.tree_to_uidef()
            target = self.tree.get_widget_id(tree_item)
            target_class = self.tree.get_widget_class(tree_item)
            class_name = self.classnamevar.get()
            
            main_widget_is_toplevel = False
            if target_class == 'tk.Toplevel':
                main_widget_is_toplevel = True
            
            context = {
                'project_name': self.projectname,
                'class_name': class_name,
                'main_widget_is_toplevel': main_widget_is_toplevel,
                'main_widget': target,
                'widget_base_class': target_class,
                'widget_code': None,
                'import_lines': None,
                'callbacks': '',
                'tkvariables': [],
            }
            
            if template == 'application':
                code = generator.generate(uidef, target, as_class=False, tabspaces=8)
                context['callbacks'] = code['callbacks']
                if self.import_tkvars_var.get():
                    context['tkvariables'] = code['tkvariables']
                tpl = makolookup.get_template('app.py.mako')
                final_code = tpl.render(**context)
                self.set_code(final_code)
            elif template == 'widget':
                code = generator.generate_widget_class(uidef, target)
                context['widget_code'] = code[target]
                context['import_lines'] = code['imports']
                context['callbacks'] = code['callbacks']
                tpl = makolookup.get_template('widget.py.mako')
                final_code = tpl.render(**context)
                self.set_code(final_code)
            elif template == 'codescript':
                code = generator.generate(uidef, target, as_class=True, 
                                          tabspaces=8)
                context['widget_code'] = code[target]
                context['import_lines'] = code['imports']
                context['callbacks'] = code['callbacks']
                tpl = makolookup.get_template('script.py.mako')
                final_code = tpl.render(**context)
                self.set_code(final_code)
    
    def on_code_copy_clicked(self):
        text = self.get_code()
        self.txt_code.clipboard_clear()
        self.txt_code.clipboard_append(text)
        
        _ = self.app.translator
        msg = _('Code copied')
        messagebox.showinfo(title=self.msgtitle, message=msg,
                            parent=self.widgetlist.winfo_toplevel())
    
    def on_code_template_changed(self, clear_code=True):
        template = self.template_var.get()
        classname = self.get_classname()
        self.cb_import_tkvars.configure(state="disabled")
        if template == 'application':
            name = '{0}App'.format(classname)
            self.classnamevar.set(name)
            self.cb_import_tkvars.configure(state="normal")
        elif template == 'codescript':
            name = '{0}App'.format(classname)
            self.classnamevar.set(name)
        elif template == 'widget':
            name = '{0}Widget'.format(classname)
            self.classnamevar.set(name)
        # Update template description
        self.template_desc_var.set(self.template_desc[template])
        if clear_code:
            self.set_code('')
    
    def on_code_save_clicked(self):
        _ = self.app.translator
        filename = (self.classnamevar.get()).lower()
        options = {
            'defaultextension': '.py',
            'filetypes': ((_('Python Script'), '*.py'), (_('All'), '*.*')),
            'initialfile': '{0}.py'.format(filename),
        }
        fname = filedialog.asksaveasfilename(**options)
        if fname:
            with open(fname, 'w') as out:
                out.write(self.get_code())
    
    def configure(self):
        self.projectname = self.app.project_name()
        wlist = self.tree.get_topwidget_list()
        self.widgetlist.configure(values=wlist)
        self.classnamevar.set(self.get_classname())
        
        if len(wlist) > 0:
            key = wlist[0][0]
            self.widgetlist_keyvar.set(key)
        self.on_code_template_changed(False)
        
    def get_classname(self):
        name = os.path.splitext(self.projectname)[0]
        return self.camel_case(name)
    
    def form_valid(self):
        valid = True
        
        _ = self.app.translator
        mbtitle = self.msgtitle
        widget = self.widgetlist.current()
        parent = self.widgetlist.winfo_toplevel()
        if widget is None:
            valid = False
            messagebox.showwarning(title=mbtitle, message=_('Select widget'),
                                   parent=parent)
        template = self.template_var.get()
        if valid and template is None:
            valid = False
            messagebox.showwarning(title=mbtitle, message=_('Select template'),
                                   parent=parent)
        classname = self.classnamevar.get()
        if valid and classname == '':
            valid = False
            messagebox.showwarning(title=mbtitle, message=_('Enter classname'),
                                   parent=parent)
        
        return valid
    
    def set_code(self, text):
        self.txt_code.delete('0.0', 'end')
        self.txt_code.insert('0.0', text)
    
    def get_code(self):
        return self.txt_code.get('0.0', 'end')

    def reset(self):
        self.set_code('')
        self.configure()
    



# encoding: UTF-8
import os
import pygubu
from pygubudesigner.i18n import translator

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
ASK_SAVE_CHANGES_DIALOG_UI = os.path.join(FILE_PATH, 'ui', 'ask_save_changes_dialog.ui')


class AskSaveChangesDialog:
    CANCEL = 0
    SAVE = 1
    DONTSAVE = 3
    
    def __init__(self, master):
        self.master = master
        self.builder = builder = pygubu.Builder(translator)
        #builder.add_resource_path(PROJECT_PATH)
        builder.add_from_file(ASK_SAVE_CHANGES_DIALOG_UI)
        self.dialog = builder.get_object('ask_save_changes_dialog', master)
        self.lbl_message = builder.get_object('lbl_message')
        self.lbl_detail = builder.get_object('lbl_detail')
        self.btn_save = builder.get_object('btn_save')
        builder.connect_callbacks(self)
        
        self.user_choice = None
    
    def on_dontsave(self):
        self.user_choice = self.DONTSAVE
        self.dialog.close()
        self.dialog.destroy()

    def on_cancel(self):
        self.user_choice = self.CANCEL
        self.dialog.close()
        self.dialog.destroy()

    def on_save(self):
        self.user_choice = self.SAVE
        self.dialog.close()
        self.dialog.destroy()

    def on_dialog_close(self, event=None):
        self.user_choice = self.CANCEL
        self.dialog.close()
        self.dialog.destroy()

    def run(self, title, message, detail=''):
        self.dialog.set_title(title)
        self.lbl_message['text'] = message
        self.lbl_detail['text'] = detail
        self.btn_save.focus_set()
        self.dialog.run()
        self.master.wait_window(self.dialog.toplevel)
        return self.user_choice


def ask_save_changes(master, title, message, detail=''):
    dialog = AskSaveChangesDialog(master)
    return dialog.run(title, message, detail)


if __name__ == '__main__':
    import tkinter as tk
    root = tk.Tk()
    app = AskSaveChangesDialog(root)
    app.run()




# encoding: utf-8
import os
import sys
import logging
from pygubudesigner.util import get_ttk_style

try:
    import tkinter as tk
    from tkinter import ttk
    from tkinter import filedialog
    from tkinter import messagebox
except:
    import Tkinter as tk
    import ttk
    import tkMessageBox as messagebox
    import tkFileDialog as filedialog
try:
    import ConfigParser as configparser
except:
    import configparser

has_appdir = False
try:
    from appdirs import AppDirs
    has_appdir = True
except:
    pass

import pygubu

logger = logging.getLogger(__name__)
FILE_PATH = os.path.dirname(os.path.abspath(__file__))
CONFIG_FILE = os.path.join(FILE_PATH, 'config')
if has_appdir:
    dirs = AppDirs('pygubu-designer')
    CONFIG_FILE = os.path.join(dirs.user_data_dir, 'config')
logger.info('Using configfile: %s', CONFIG_FILE)

options = {
    'widget_set': {'values': '["tk", "ttk"]', 'default':'ttk'},
    'ttk_theme': {'default': 'default'},
    'geometry': {
        'default': '640x480',
        },
    'widget_naming_separator': {
        'values': '["NONE", "UNDERSCORE"]',
        'default': 'NONE',
        },
    'widget_naming_ufletter': {
        'values': '["yes", "no"]',
        'default': 'no',
        },
    }

SEC_GENERAL = 'GENERAL'
SEC_CUSTOM_WIDGETS = 'CUSTOM_WIDGETS'
SEC_RECENT_FILES = 'RECENT_FILES'
config = configparser.SafeConfigParser()
config.add_section(SEC_CUSTOM_WIDGETS)
config.add_section(SEC_GENERAL)
config.add_section(SEC_RECENT_FILES)


def initialize_configfile():
    if not os.path.exists(dirs.user_data_dir):
        os.makedirs(dirs.user_data_dir)
    if not os.path.exists(CONFIG_FILE):
        with open(CONFIG_FILE, 'w') as configfile:
            config.write(configfile)

def save_configfile():
    with open(CONFIG_FILE, 'w') as configfile:
        config.write(configfile)

def load_configfile():
    defaults = {}
    for k in options:
        defaults[k] = options[k]['default']
    if sys.version_info < (3,0):
        #Python 2.7
        keys = defaults.keys()
        for i in range(0, len(defaults)):
            config.set(SEC_GENERAL, keys[i], defaults.get(keys[i]))
    else:
        #Python 3
        config[SEC_GENERAL] = defaults
    if not os.path.exists(CONFIG_FILE):
        initialize_configfile()
    else:
        try:
            config.read(CONFIG_FILE)
        except configparser.MissingSectionHeaderError as e:
            logger.exception(e)
            msg = _("Configuration file at '{0}' is corrupted, program may not work as expected.\nIf you delete this file, configuration will be set to default".format(CONFIG_FILE))
            messagebox.showerror(_('Error'), msg)
        except configparser.Error as e:
            logger.exception(e)
            msg = _("Faild to parse config file at '{0}', program may not work as expected.".format(CONFIG_FILE))
            msg = msg.format(CONFIG_FILE)
            messagebox.showerror(_('Error'), msg)

def get_custom_widgets():
    paths = []
    for k, p in config.items(SEC_CUSTOM_WIDGETS):
        paths.append(p)
    return paths
    
def get_option(key):
    return config.get(SEC_GENERAL, key)

def set_option(key, value, save=False):
    config.set(SEC_GENERAL, key, value)
    if save:
        save_configfile()

def recent_files_get():
    rf = []
    for k, f in config.items(SEC_RECENT_FILES):
        rf.append(f)
    return rf

def recent_files_save(file_list):
    config.remove_section(SEC_RECENT_FILES)
    config.add_section(SEC_RECENT_FILES)
    for j, p in enumerate(file_list):
        config.set(SEC_RECENT_FILES, 'f{0}'.format(j), p)
    save_configfile()

def save_window_size(geom):
    set_option('geometry', geom, save=True)

def get_window_size():
    return get_option('geometry')

# Get user configuration
load_configfile()


class PreferencesUI(object):

    def __init__(self, master, translator=None):
        self.master = master
        self.translator = translator
        self.dialog = None
        self.builder = None
        self._create_preferences_dialog()
        self._load_options()

    def _create_preferences_dialog(self):
        self.builder = builder = pygubu.Builder(self.translator)
        uifile = os.path.join(FILE_PATH, "ui/preferences_dialog.ui")
        builder.add_from_file(uifile)

        top = self.master.winfo_toplevel()
        self.dialog = dialog = builder.get_object('preferences', top)
        
        # setup theme values
        s = get_ttk_style()
        styles = s.theme_names()
        themelist = []
        for name in styles:
            themelist.append((name, name))
        themelist.sort()
        cbox = builder.get_object('cbox_ttk_theme')
        cbox.configure(values=themelist)
        
        #General
        for key in ('widget_set',):
            cbox = builder.get_object(key)
            cbox.configure(values=options[key]['values'])
        
        #Custom widgets
        self.cwtv = builder.get_object('cwtv')
        self.path_remove = builder.get_object('path_remove')
        builder.connect_callbacks(self)
        
        # hide options
        fhide = builder.get_object('fhidden')
        fhide.pack_forget()
        
    def _load_options(self):
        # General
        for key in options:
            var = self.builder.get_variable(key)
            var.set(get_option(key))
        # Custom widgets
        for k, p in config.items(SEC_CUSTOM_WIDGETS):
            self.cwtv.insert('', 'end', text=p)
        self._configure_path_remove()

    def _save_options(self):
        # General
        for key in options:
            var = self.builder.get_variable(key)
            config.set(SEC_GENERAL, key, var.get())
        # Custom Widgets
        config.remove_section(SEC_CUSTOM_WIDGETS)
        config.add_section(SEC_CUSTOM_WIDGETS)
        paths = []
        for iid in self.cwtv.get_children():
            txt = self.cwtv.item(iid, 'text')
            paths.append(txt)
        for j, p in enumerate(paths):
            config.set(SEC_CUSTOM_WIDGETS, 'w{0}'.format(j), p)
        save_configfile()
        self.master.event_generate('<<PygubuDesignerPreferencesSaved>>')

    def _configure_path_remove(self):
        if len(self.cwtv.get_children()):
            self.path_remove.configure(state='normal')
        else:
            self.path_remove.configure(state='disabled')
        
    def on_dialog_close(self, event=None):
        self._save_options()
        self.dialog.close()

    def on_pathremove_clicked(self):
        items = self.cwtv.selection()
        self.cwtv.delete(*items)
        self._configure_path_remove()

    def on_pathadd_clicked(self):
        options = {
            'defaultextension': '.py',
            'filetypes': ((_('Python module'), '*.py'), (_('All'), '*.*'))}
        fname = filedialog.askopenfilename(**options)
        if fname:
            self.cwtv.insert('', tk.END, text=fname)
            self._configure_path_remove()
        



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here
from __future__ import unicode_literals
import logging

from pygubu.builder import CLASS_MAP
from pygubu.builder.widgetmeta import WidgetMeta as WidgetMetaBase, BindingMeta
from .util.observable import Observable
from .properties import (WIDGET_PROPERTIES, PACK_PROPERTIES, PLACE_PROPERTIES,
                         GRID_PROPERTIES, LAYOUT_OPTIONS)

logger = logging.getLogger(__name__)


class WidgetMeta(WidgetMetaBase, Observable):
    
    def apply_layout_defaults(self):
        super(WidgetMeta, self).apply_layout_defaults()
        self.notify('LAYOUT_CHANGED', self)
        
    def widget_property(self, name, value=None):
        if value is None:
            if name == 'id':
                return self.identifier
            elif name == 'class':
                return self.classname
            else:
                return self.properties.get(name, '')
        else:
            # Setter
            if name == 'id':
                self.identifier = value
            elif name == 'class':
                self.classname = value
            else:
                if value:
                    self.properties[name] = value
                else:
                    # remove if no value set
                    self.properties.pop(name, None)
            self.notify('PROPERTY_CHANGED', self)
    
    def layout_property(self, name, value=None):
        if value is None:
            # Getter
            default = ''
            if name in ('row', 'column'):
                default = '0'
            return self.layout_properties.get(name, default)
        else:
            # Setter
            if value:
                self.layout_properties[name] = value
            else:
                # remove if no value set
                self.layout_properties.pop(name, None)
            self.notify('LAYOUT_CHANGED', self)
    
    def gridrc_property(self, type_, num, pname, value=None):
        if value is None:
            # Getter
            default = '0'
            if pname == 'uniform':
                default = ''
            pvalue = self.get_gridrc_value(type_, num, pname)
            pvalue = pvalue if pvalue is not None else default
            return pvalue
        else:
            # Setter
            self.set_gridrc_value(type_, num, pname, value)
            self.notify('LAYOUT_GRIDRC_CHANGED', self)

    
    @property
    def manager(self):
        return self._manager
    
    @manager.setter
    def manager(self, value):
        if self._manager != value:
            self._manager = value
            self.clear_layout()
        self.notify('LAYOUT_CHANGED', self)

    def get_bindings(self):
        blist = []
        for v in self.bindings:
            blist.append((v.sequence, v.handler, v.add))
        return blist

    def clear_bindings(self):
        self.bindings = []

    def add_binding(self, seq, handler, add):
        self.bindings.append(BindingMeta(seq, handler, add))

    def remove_unused_grid_rc(self):
        """Deletes unused grid row/cols"""
        pass
    
    def setup_defaults(self):
        propd, layoutd = WidgetMeta.get_widget_defaults(self, self.identifier)
        self.properties_defaults = propd
        self.layout_defaults = layoutd
    
    @staticmethod
    def get_widget_defaults(wclass, widget_id):
        properties = {}
        layout = {}
        
        if wclass in CLASS_MAP:
            # setup default values for properties
            for pname in CLASS_MAP[wclass].builder.properties:
                pdescription = {}
                if pname in WIDGET_PROPERTIES:
                    pdescription = WIDGET_PROPERTIES[pname]
                if wclass in pdescription:
                    pdescription = dict(pdescription, **pdescription[wclass])
                default_value = str(pdescription.get('default', ''))
                if default_value:
                    properties[pname] = default_value
                # default text for widgets with text prop:
                if pname in ('text', 'label'):
                    properties[pname] = widget_id
        
        # setup default values for layout
        groups = (
            ('pack', PACK_PROPERTIES),
            ('place', PLACE_PROPERTIES),
            ('grid', GRID_PROPERTIES),
        )
        for manager, manager_prop in groups:
            layout[manager] = {}
            for pname in manager_prop:
                pdescr = LAYOUT_OPTIONS[pname]
                if manager in pdescr:
                    pdescr = pdescr.copy()
                    pdescr = dict(pdescr, **pdescr[manager])
                default_value = pdescr.get('default', None)
                if default_value:
                    layout[manager][pname] = default_value
        return properties, layout

